
# Copyright 2013-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module defines the L{main()} function for executing the zim
application. It also defines a number of command classes that implement
specific commandline commands and an singleton application object that
takes core of the process life cycle.
'''

import logging

logger = logging.getLogger('zim')

from contextlib import contextmanager

import zim
import zim.newfs
import zim.config

from zim import __version__

from zim.base.klasslookup import get_module, lookup_subclass
from zim.errors import Error
from zim.notebook import Path, HRef, \
	get_notebook_list, resolve_notebook, build_notebook
from zim.formats import get_format

from zim.config import ConfigManager
from zim.plugins import PluginManager

from .command import Command, GtkCommand, UsageError, GetoptError
	# Keep GetOptError import - not used here, but imported from here by zim.py


class HelpCommand(Command):
	'''Class implementing the C{--help} command'''

	usagehelp = '''\
usage: zim [OPTIONS] [NOTEBOOK [PAGE_LINK]]
   or: zim --gui [OPTIONS] [NOTEBOOK [PAGE_LINK]]
   or: zim --server [OPTIONS] [NOTEBOOK]
   or: zim --export [OPTIONS] NOTEBOOK [PAGE]
   or: zim --import [OPTIONS] NOTEBOOK PAGE FILES
   or: zim --search [OPTIONS] NOTEBOOK QUERY
   or: zim --index  [OPTIONS] NOTEBOOK
   or: zim --plugin PLUGIN [ARGUMENTS]
   or: zim --manual [OPTIONS] [PAGE_LINK]
   or: zim --help

NOTEBOOK can be a local file path, a local file URI or a notebook name
PAGE is be a fully specified page name
PAGE_LINK is a fully specified page name optionally extended with an anchor ID
'''
	optionhelp = '''\
General Options:
  --gui             run the editor (this is the default)
  --server          run the web server
  --export          export to a different format
  --import          import one or more files into a notebook
  --search          run a search query on a notebook
  --index           build an index for a notebook
  --plugin          call a specific plugin function
  --manual          open the user manual
  -V, --verbose     print information to terminal
  -D, --debug       print debug messages
  -v, --version     print version and exit
  -h, --help        print this text

GUI Options:
  --list            show the list with notebooks instead of
                    opening the default notebook
  --geometry        window size and position as WxH+X+Y
  --fullscreen      start in fullscreen mode
  --non-unique      start a new process, do not connect to an existing process
  --standalone      start a new process per notebook, implies --non-unique

Server Options:
  --port            port to use (defaults to 8080)
  --template        name or filepath of the template to use
  --private         serve only to localhost
  --gui             run the gui wrapper for the server

Export Options:
  -o, --output      output directory (mandatory option)
  --format          format to use (defaults to 'html')
  --template        name or filepath of the template to use
  --root-url        url to use for the document root
  --index-page      index page name
  -r, --recursive   when exporting a page, also export sub-pages
  -s, --singlefile  export all pages to a single output file
  -O, --overwrite   force overwriting existing file(s)

Import Options:
  --format          format to read (defaults to 'wiki')
  --assubpage       import files as sub-pages of PATH, this is implicit true
                    when PATH ends with a ":" or when multiple files are given

Search Options:
  -s, --with-scores print score for each page, sort by score

Index Options:
  -f, --flush       flush the index first and force re-building

Try 'zim --manual' for more help.
'''

	def __init__(self, command, cmdhelp=None, pwd=None):
		Command.__init__(self, command, pwd=None)
		self.cmdhelp = cmdhelp

	def run(self):
		if self.cmdhelp:
			print(self.cmdhelp)
		else:
			print(self.usagehelp)
			print(self.optionhelp)  # TODO - generate from commands


class VersionCommand(Command):
	'''Class implementing the C{--version} command'''

	def run(self):
		print('zim %s\n' % zim.__version__)
		print(zim.__copyright__, '\n')
		print(zim.__license__)


class NotebookLookupError(Error):
	'''Error when failing to locate a notebook'''

	description = _('Could not find the file or folder for this notebook')
		# T: Error verbose description


class NotebookCommand(Command):
	'''Base class for commands that act on a notebook'''

	def get_default_or_only_notebook(self):
		'''Helper to get a default notebook'''
		notebooks = get_notebook_list()
		if notebooks.default:
			uri = notebooks.default.uri
		elif len(notebooks) == 1:
			uri = notebooks[0].uri
		else:
			return None

		return resolve_notebook(uri, pwd=self.pwd) # None if not found

	def get_notebook_argument(self):
		'''Get the notebook and page arguments for this command
		@returns: a 2-tuple of an L{NotebookInfo} object and an
		optional L{HRef} or C{(None, None)} if the notebook
		argument is optional and not given
		@raises NotebookLookupError: if the notebook is mandatory and
		not given, or if it is given but could not be resolved
		'''
		assert self.arguments[0] in ('NOTEBOOK', '[NOTEBOOK]')
		args = self.get_arguments()
		notebook = args[0]

		if notebook is None:
			if self.arguments[0] == 'NOTEBOOK': # not optional
				raise NotebookLookupError(_('Please specify a notebook'))
					# T: Error when looking up a notebook
			else:
				return None, None

		notebookinfo = resolve_notebook(notebook, pwd=self.pwd)
		if not notebookinfo:
			raise NotebookLookupError(_('Could not find notebook: %s') % notebook)
				# T: error message

		if len(self.arguments) > 1 \
		and self.arguments[1] in ('PAGE', '[PAGE]', 'PAGE_LINK', '[PAGE_LINK]') \
		and args[1] is not None and args[1] != ':':
			# The ":" path is a special case to allow e.g. --import to handle top level namespace
			return notebookinfo, HRef.new_from_wiki_link(args[1])
		else:
			return notebookinfo, None

	def build_notebook(self, ensure_uptodate=True):
		'''Get the L{Notebook} object for this command
		Tries to automount the file location if needed.
		@param ensure_uptodate: if C{True} index is updated when needed.
		Only set to C{False} when index update is handled explicitly
		(e.g. in the main gui).
		@returns: a L{Notebook} object and a C{HRef} object or C{None}
		@raises NotebookLookupError: if the notebook could not be
		resolved or is not given
		@raises FileNotFoundError: if the notebook location does not
		exist and could not be mounted.
		'''
		# Explicit page argument has priority over implicit from uri
		# mounting is attempted by zim.notebook.build_notebook()
		notebookinfo, pagelink = self.get_notebook_argument() 	# can raise NotebookLookupError
		if not notebookinfo:
			raise NotebookLookupError(_('Please specify a notebook'))
		notebook, uripagelink = build_notebook(notebookinfo) # can raise FileNotFoundError

		if ensure_uptodate and not notebook.index.is_uptodate:
			for info in notebook.index.update_iter():
				#logger.info('Indexing %s', info)
				pass # TODO meaningful info for above message

		return notebook, pagelink or uripagelink


class GuiCommand(NotebookCommand, GtkCommand):
	'''Class implementing the C{--gui} command and run the gtk interface'''

	arguments = ('[NOTEBOOK]', '[PAGE_LINK]')
	options = (
		('list', '', 'show the list with notebooks instead of\nopening the default notebook'),
		('geometry=', '', 'window size and position as WxH+X+Y'),
		('fullscreen', '', 'start in fullscreen mode'),
		('non-unique', '', 'start a new process, do not connect to an existing process'),
		('standalone', '', 'start a new process per notebook, implies --non-unique'),
	)

	def build_notebook(self, ensure_uptodate=False):
		# Bit more complicated here due to options to use default and
		# allow using notebookdialog to prompt

		# Explicit page argument has priority over implicit from uri
		# mounting is attempted by zim.notebook.build_notebook()

		from zim.notebook import FileNotFoundError

		def prompt_notebook_list():
			import zim.gui.notebookdialog
			return zim.gui.notebookdialog.prompt_notebook()
				# Can return None if dialog is cancelled

		used_default = False
		pagelink, uripagelink = None, None
		if self.opts.get('list'):
			notebookinfo = prompt_notebook_list()
		else:
			notebookinfo, pagelink = self.get_notebook_argument()

			if notebookinfo is None:
				notebookinfo = self.get_default_or_only_notebook()
				used_default = notebookinfo is not None

				if notebookinfo is None:
					notebookinfo = prompt_notebook_list()

		if notebookinfo is None:
			return None, None # Cancelled prompt

		try:
			notebook, uripagelink = build_notebook(notebookinfo) # can raise FileNotFound
		except FileNotFoundError:
			if used_default:
				# Default notebook went missing? Fallback to dialog to allow changing it
				notebookinfo = prompt_notebook_list()
				if notebookinfo is None:
					return None, None # Cancelled prompt
				notebook, uripagelink = build_notebook(notebookinfo) # can raise FileNotFound
			else:
				raise

		if ensure_uptodate and not notebook.index.is_uptodate:
			for info in notebook.index.update_iter():
				#logger.info('Indexing %s', info)
				pass # TODO meaningful info for above message

		return notebook, pagelink or uripagelink

	def run(self):
		from gi.repository import Gtk

		from zim.gui.mainwindow import MainWindow

		windows = [
			w for w in Gtk.Window.list_toplevels()
				if isinstance(w, MainWindow)
		]

		notebook, pagelink = self.build_notebook()
		if notebook is None:
			logger.debug('NotebookDialog cancelled - exit')
			return

		for window in windows:
			if window.notebook.uri == notebook.uri:
				self._present_window(window, pagelink)
				return window
		else:
			return self._run_new_window(notebook, pagelink)

	def _present_window(self, window, pagelink):
		window.present()

		if pagelink:
			window.open_page(Path(pagelink.names), anchor=pagelink.anchor)

		geometry = self.opts.get('geometry', None)
		if geometry is not None:
			window.parse_geometry(geometry)

		if self.opts.get('fullscreen', False):
			window.toggle_fullscreen(True)

	def _run_new_window(self, notebook, pagelink):
		from gi.repository import GObject

		from zim.gui.mainwindow import MainWindow

		pluginmanager = PluginManager()

		preferences = ConfigManager.preferences['General']
		preferences.setdefault('plugins_list_version', 'none')
		if preferences['plugins_list_version'] != '0.70':
			if not preferences['plugins']:
				pluginmanager.load_plugins_from_preferences(
					[ # Default plugins
						'pageindex', 'pathbar', 'toolbar',
						'insertsymbol', 'printtobrowser',
						'versioncontrol', 'osx_menubar'
					]
				)
			else:
				# Upgrade version <0.70 where these were core functions
				pluginmanager.load_plugins_from_preferences(['pageindex', 'pathbar'])

			if 'calendar' in pluginmanager.failed:
				ConfigManager.preferences['JournalPlugin'] = \
						ConfigManager.preferences['CalendarPlugin']
				pluginmanager.load_plugins_from_preferences(['journal'])

			preferences['plugins_list_version'] = '0.70'

		page = Path(pagelink.names) if pagelink else None
		window = MainWindow(
			notebook,
			page=page,
			**self.get_options('geometry', 'fullscreen')
		)
		window.present()
		if pagelink and pagelink.anchor:
			window.open_page(Path(pagelink.names), anchor=pagelink.anchor)

		if not window.notebook.index.is_uptodate:
			window._uiactions.check_and_update_index(update_only=True) # XXX
		else:
			# Start a lightweight background check of the index
			# put a small delay to ensure window is shown before we start
			def start_background_check():
				notebook.index.start_background_check(notebook)
				return False # only run once
			GObject.timeout_add(500, start_background_check)

		return window


class ManualCommand(GuiCommand):
	'''Like L{GuiCommand} but always opens the manual'''

	arguments = ('[PAGE_LINK]',)
	options = tuple(t for t in GuiCommand.options if t[0] != 'list')
		# exclude --list

	def run(self):
		from zim.config import data_dir
		self.arguments = ('NOTEBOOK', '[PAGE_LINK]') # HACK
		self.args.insert(0, data_dir('manual').path)
		return GuiCommand.run(self)


class ServerCommand(NotebookCommand):
	'''Class implementing the C{--server} command and running the web
	server.
	'''

	arguments = ('NOTEBOOK',)
	options = (
		('port=', 'p', 'port number to use (defaults to 8080)'),
		('template=', 't', 'name or path of the template to use'),
		('private', '', 'serve only to localhost')
	)

	def run(self):
		import zim.www
		from zim.templates import get_template

		port = int(self.opts.get('port', 8080))
		template = get_template('html', self.opts.get('template', 'Default'), pwd=self.pwd)
		notebook, x = self.build_notebook()
		is_public = not self.opts.get('private', False)

		self.server = httpd = zim.www.make_server(notebook, public=is_public, template=template, port=port)
			# server attribute used in testing to stop sever in thread
		logger.info("Serving HTTP on %s port %i...", httpd.server_name, httpd.server_port)
		httpd.serve_forever()


class ServerGuiCommand(NotebookCommand, GtkCommand):
	'''Like L{ServerCommand} but uses the graphical interface for the
	server defined in L{zim.gui.server}.
	'''

	arguments = ('[NOTEBOOK]',)
	options = (
		('port=', 'p', 'port number to use (defaults to 8080)'),
		('template=', 't', 'name or path of the template to use'),
		('non-unique', '', 'start a new process, do not connect to an existing process'),
	)

	def run(self):
		import zim.gui.server
		from zim.templates import get_template

		port = int(self.opts.get('port', 8080))
		template = get_template('html', self.opts.get('template', 'Default'), pwd=self.pwd)
		notebookinfo, x = self.get_notebook_argument()
		if notebookinfo is None:
			# Prefer default to be selected in drop down, user can still change
			notebookinfo = self.get_default_or_only_notebook()

		window = zim.gui.server.ServerWindow(
			notebookinfo,
			public=True,
			template=template,
			port=port
		)
		window.show_all()
		return window


class ExportCommand(NotebookCommand):
	'''Class implementing the C{--export} command'''

	arguments = ('NOTEBOOK', '[PAGE]')
	options = (
		('format=', '', 'format to use (defaults to \'html\')'),
		('template=', '', 'name or path of the template to use'),
		('output=', 'o', 'output folder, or output file name'),
		('root-url=', '', 'url to use for the document root'),
		('index-page=', '', 'index page name'),
		('recursive', 'r', 'when exporting a page, also export sub-pages'),
		('singlefile', 's', 'export all pages to a single output file'),
		('overwrite', 'O', 'overwrite existing file(s)'),
	)

	def get_exporter(self, page):
		from zim.newfs import localFileOrFolder, FilePath, LocalFile, LocalFolder, FileNotFoundError
		from zim.templates import get_template
		from zim.export import \
			build_mhtml_file_exporter, \
			build_single_file_exporter, \
			build_page_exporter, \
			build_notebook_exporter

		format = self.opts.get('format', 'html')
		template = get_template(format, self.opts.get('template', 'Default'), pwd=self.pwd)

		if not 'output' in self.opts:
			raise UsageError(_('Output location needed for export')) # T: error in export command

		try:
			output = localFileOrFolder(self.opts['output'], pwd=self.pwd)
		except FileNotFoundError:
			# resolve path, but type undecided
			output = FilePath(self.pwd).get_abspath(self.opts['output']) # can raise again for mal-formed paths
		else:
			# file or folder exists
			if not self.opts.get('overwrite'):
				if isinstance(output, LocalFolder):
					if len(output.list_names()) > 0:
						raise Error(_('Output folder exists and not empty, specify "--overwrite" to force export'))  # T: error message for export
					else:
						pass
				else:
					raise Error(_('Output file exists, specify "--overwrite" to force export'))  # T: error message for export

		if format == 'mhtml':
			self.ignore_options('index-page')
			if isinstance(output, LocalFolder): # implies exists
				raise UsageError(_('Need output file to export MHTML')) # T: error message for export
			else:
				output = LocalFile(output) # FilePath --> LocalFile

			exporter = build_mhtml_file_exporter(
				output, template,
				document_root_url=self.opts.get('root-url'),
			)
		elif self.opts.get('singlefile'):
			self.ignore_options('index-page')
			if isinstance(output, LocalFolder):
				ext = get_format(format).info['extension']
				output = output.file(page.basename) + '.' + ext
			else:
				output = LocalFile(output) # FilePath --> LocalFile

			exporter = build_single_file_exporter(
				output, format, template, namespace=page,
				document_root_url=self.opts.get('root-url'),
			)
		elif page:
			self.ignore_options('index-page')
			if isinstance(output, LocalFolder):
				ext = get_format(format).info['extension']
				output = output.file(page.basename) + '.' + ext
			else:
				output = LocalFile(output) # FilePath --> LocalFile

			exporter = build_page_exporter(
				output, format, template, page,
				document_root_url=self.opts.get('root-url'),
			)
		else:
			if isinstance(output, LocalFile): # implies exists
				raise UsageError(_('Need output folder to export full notebook')) # T: error message for export
			else:
				output = LocalFolder(output) # FilePath --> LocalFolder

			exporter = build_notebook_exporter(
				output, format, template,
				index_page=self.opts.get('index-page'),
				document_root_url=self.opts.get('root-url'),
			)

		return exporter

	def run(self):
		from zim.export.selections import AllPages, SinglePage, SubPages

		notebook, href = self.build_notebook()

		if href and self.opts.get('recursive'):
			page = Path(href.names) # ignore anchor
			selection = SubPages(notebook, page)
		elif href:
			page = Path(href.names) # ignore anchor
			selection = SinglePage(notebook, page)
		else:
			page = None
			selection = AllPages(notebook)

		exporter = self.get_exporter(page)
		exporter.export(selection)


class ImportCommand(NotebookCommand):
	'''Class implementing the C{--import} command'''

	arguments = ('NOTEBOOK', 'PAGE', 'FILES+')
	options = (
		('format=', '', 'format to import from (defaults to \'wiki\')'),
		('assubpage', 's', 'import as sub-pages of PATH'),
	)

	def run(self):
		from zim.newfs import localFileOrFolder
		from zim.import_files import import_file_from_user_input, import_files_from_user_input

		notebook, href = self.build_notebook()
		path = Path(href.names) if href else None
		format = self.opts.get('format', 'wiki')
		assubpage = self.opts.get('assubpage', False)

		n, p, *files = self.get_arguments()
		files = [localFileOrFolder(f, pwd=self.pwd) for f in files] # raises if does not exist

		if p.endswith(':') or len(files) > 1:
			# Implicit set option, only real use of the option is to force single file imports
			# as sub-page instead of target page
			assubpage = True

		if not assubpage:
			# Special case for 1 file to 1 page
			if notebook.get_page(path).exists():
				raise UsageError('Page "%s" exists, to import as sub-page please use "--assubpage"' % path.name)
			import_file_from_user_input(files[0], notebook, path, format)
		else:
			import_files_from_user_input(files, notebook, path, format)


class SearchCommand(NotebookCommand):
	'''Class implementing the C{--search} command'''

	arguments = ('NOTEBOOK', 'QUERY')
	options = (
		("with-scores", "s", "also print scores of search results"),
	)

	def run(self):
		from zim.search import SearchSelection, Query

		notebook, x = self.build_notebook()
		n, query = self.get_arguments()

		if query and not query.isspace():
			logger.info('Searching for: %s', query)
			query = Query(query)
		else:
			raise ValueError('Empty query')

		selection = SearchSelection(notebook)
		selection.search(query)

		if self.opts.get("with-scores", False):
			sorted_sel = sorted(selection.scores.items(),
				key=lambda i:i[0].name, reverse=False)
			sorted_sel.sort(key=lambda i:i[1], reverse=True)

			for result in sorted_sel:
				print(str(result[1]) + "\t" + result[0].name)
		else:
			for path in sorted(selection, key=lambda p: p.name):
				print(path.name)


class IndexCommand(NotebookCommand):
	'''Class implementing the C{--index} command'''

	arguments = ('NOTEBOOK',)
	options = (
		('flush', 'f', 'flush the index first and force re-building'),
	)

	def run(self):
		# Elevate logging level of indexer to ensure "zim --index -V" gives
		# some meaningfull output
		def elevate_index_logging(log_record):
			if log_record.levelno == logging.DEBUG:
				log_record.levelno = logging.INFO
				log_record.levelname = 'INFO'
			return True

		mylogger = logging.getLogger('zim.notebook.index')
		mylogger.setLevel(logging.DEBUG)
		mylogger.addFilter(elevate_index_logging)

		notebook, x = self.build_notebook(ensure_uptodate=False)
		if self.opts.get('flush'):
			notebook.index.flush()
			notebook.index.update()
		else:
			# Effectively the same as check_and_update_index ui action
			logger.info('Checking notebook index')
			notebook.index.check_and_update()

		logger.info('Index up to date!')


commands = {
	'help': HelpCommand,
	'version': VersionCommand,
	'gui': GuiCommand,
	'manual': ManualCommand,
	'server': ServerCommand,
	'servergui': ServerGuiCommand,
	'export': ExportCommand,
	'import': ImportCommand,
	'search': SearchCommand,
	'index': IndexCommand,
}


def build_command(args, pwd=None):
	'''Parse all commandline options
	@param args: commandline argumnets, starting with first switch or option
	@param pwd: working directory
	@returns: a L{Command} object
	@raises UsageError: if args is not correct
	'''
	args = list(args)
	if args and args[0] == '--plugin':
		args.pop(0)
		try:
			cmd = args.pop(0)
		except IndexError:
			raise UsageError('Missing plugin name')

		try:
			mod = get_module('zim.plugins.' + cmd)
			klass = lookup_subclass(mod, Command)
		except:
			if '-D' in args or '--debug' in args:
				logger.exception('Error while loading: zim.plugins.%s.Command', cmd)
				# Can't use following because log level not yet set:
				# logger.debug('Error while loading: zim.plugins.%s.Command', cmd, exc_info=sys.exc_info())
			raise UsageError('Could not load commandline command for plugin "%s"' % cmd)
		if klass is None:
			raise UsageError('Module %s has no commandline command' % cmd)
	else:
		if args and args[0].startswith('--') and args[0][2:] in commands:
			cmd = args.pop(0)[2:]
			if cmd == 'server' and '--gui' in args:
				args.remove('--gui')
				cmd = 'servergui'
		elif args and args[0] == '-v':
			args.pop(0)
			cmd = 'version'
		elif args and args[0] == '-h':
			args.pop(0)
			cmd = 'help'
		else:
			cmd = 'gui' # default

		klass = commands[cmd]

	obj = klass(cmd, pwd=pwd)
	obj.parse_options(*args)

	# Hack to support --help for plugin commands in local process
	# convert to HelpCommand
	if obj.opts.get('help'):
		obj = HelpCommand('help', cmdhelp=obj.cmdhelp, pwd=pwd)
		obj.parse_options(*args)

	return obj


def _application_startup():
	# Common startup between Gtk application, and non-Gtk commands
	
	## Log application info at startup
	logger.info('This is zim %s', __version__)
	level = logger.getEffectiveLevel()
	if level == logging.DEBUG:
		import sys
		import os
		import zim.config

		logger.debug('Python version is %s', str(sys.version_info))
		logger.debug('Platform is %s', os.name)
		zim.config.log_basedirs()

	## Load plugins before any extendable objects are loaded
	PluginManager().load_plugins_from_preferences(
		ConfigManager.preferences['General']['plugins']
	)


_zim_gtk_application = None


def main(*argv):
	'''Run full zim application
	@param argv: the commandline arguments, as given by C{sys.argv}
	@returns: exit code (if error handled in application process, else just raises)
	'''
	
	# See the `main()`` function in the `zim.py` script for bootstrapping the 
	# environment before running the application.

	# The logic here is:
	# 1. Parse the commandline to determine which Command to run
	# 2. If the Command is _not_ a GtkCommand, setup logging and run the Command
	# 3. If the Command is a GtkCommand, start the application and pass on the standalone / non-unique flag
	# 4. In the application parse the commandline again to re-build the Command in the primary process and run it

	global _zim_gtk_application

	cmd = build_command(argv[1:])
	if isinstance(cmd, GtkCommand):
		from .application import ZimGtkApplication
		_zim_gtk_application = ZimGtkApplication(
			non_unique=cmd.opts.get('non-unique'),
			standalone=cmd.opts.get('standalone')
		)
		argv = [argv[0]] + list(cmd.handle_local_commandline(list(argv[1:])))
		return _zim_gtk_application.run(argv)
	else:
		_application_startup()
		return cmd.run()



#
# gnuplot_ploteditor.py
#
# This is a plugin for Zim, which allows inserting Gnuplot scripts to
# have Zim generate plots from them.
#
# Author: Alessandro Magni <magni@inrim.it>
# Date: 2010-10-12
# Copyright (c) 2010, released under the GNU GPL v2 or higher
#
#

import glob

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import \
	ImageGeneratorClass, BackwardImageGeneratorObjectType

from zim.newfs import LocalFile, TmpFile
from zim.config import data_file
from zim.templates import get_template
from zim.applications import Application, ApplicationError


# TODO put these commands in preferences
gnuplot_cmd = ('gnuplot',)


class InsertGnuplotPlugin(PluginClass):

	plugin_info = {
		'name': _('Insert Gnuplot'), # T: plugin name
		'description': _('''\
This plugin provides a plot editor for zim based on Gnuplot.
'''), # T: plugin description
		'help': 'Plugins:Gnuplot Editor',
		'author': 'Alessandro Magni',
	}

	@classmethod
	def check_dependencies(klass):
		has_gnuplot = Application(gnuplot_cmd).tryexec()
		return has_gnuplot, [('Gnuplot', has_gnuplot, True)]


class BackwardGnuplotImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+gnuplot'
	label = _('Gnuplot') # T: menu item
	syntax = None
	scriptname = 'gnuplot.gnu'


class GnuplotGenerator(ImageGeneratorClass):

	imagefile_extension = '.png'

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.template = get_template('plugins', 'gnuploteditor.gnu')
		self.attachment_folder = notebook.get_attachments_dir(page)
		self.plotscriptfile = TmpFile('gnuplot.gnu')

	def generate_image(self, text):
		plotscriptfile = self.plotscriptfile
		pngfile = LocalFile(plotscriptfile.path[:-4] + '.png')

		template_vars = { # they go in the template
			'gnuplot_script': text,
			'png_fname': pngfile.path,
		}
		if self.attachment_folder and self.attachment_folder.exists():
			template_vars['attachment_folder'] = self.attachment_folder.path
		else:
			template_vars['attachment_folder'] = ''

		# Write to tmp file using the template for the header / footer
		lines = []
		self.template.process(lines, template_vars)
		plotscriptfile.writelines(lines)
		#~ print('>>>\n%s<<<' % plotscriptfile.read())

		# Call Gnuplot
		try:
			gnu_gp = Application(gnuplot_cmd)
			gnu_gp.run(args=(plotscriptfile.basename, ), cwd=plotscriptfile.parent())
							# you call it as % gnuplot output.plt

		except ApplicationError:
			return None, None # Sorry - no log
		else:
			return pngfile, None

	def cleanup(self):
		path = self.plotscriptfile.path
		for path in glob.glob(path[:-4] + '.*'):
			LocalFile(path).remove()




# Copyright 2009-2017 Jaap Karssenberg <jaap.karssenberg@gmail.com>


import logging

logger = logging.getLogger('zim.notebook.index')

from zim.signals import SignalEmitter, ConnectorMixin


class IndexNotFoundError(ValueError):
	'''Error used when lookup fails because a pagename does not appear
	in the index.
	'''
	pass


class IndexConsistencyError(AssertionError):
	'''Error used when a lookup fails while expected to succeed'''
	pass


class IndexView(object):
	'''Base class for "index view" objects'''

	@classmethod
	def new_from_index(cls, index):
		return cls(index._db)

	def __init__(self, db):
		self.db = db


class IndexerBase(SignalEmitter, ConnectorMixin):
	'''Base class for "content indexer" objects.
	It defines the callback functions that are calls from L{PagesIndexer}
	'''

	__signals__ = {}

	def __init__(self, db):
		self.db = db

	def is_uptodate(self):
		return True

	def update(self):
		for i in self.update_iter():
			pass

	def update_iter(self):
		return iter([])


class MyTreeIter(object):
	__slots__ = ('treepath', 'row', 'n_children', 'hint')

	def __init__(self, treepath, row, n_children, hint=None):
		self.treepath = treepath
		self.row = row
		self.n_children = n_children
		self.hint = hint


class TreeModelMixinBase(ConnectorMixin):
	'''This class can be used as mixin class for C{Gtk.TreeModel}
	implementations that use data from the index.

	Treepaths are simply tuples with integers. This Mixin assumes L{MyTreeIter}
	objects for iters. (Which should not be confused with C{Gtk.TreeIter} as
	used by the interface!)
	'''

	def __init__(self, index):
		self.index = index
		self.db = index._db
		self.cache = {}
		self.connect_to_updateiter(index, index.update_iter)
		self.connectto(index, 'new-update-iter', self.connect_to_updateiter)

	def connect_to_updateiter(self, update_iter):
		'''Connect to a new L{IndexUpdateIter}

		The following signals must be implemented:

		  - row-inserted (treepath, treeiter)
		  - row-changed (treepath, treeiter)
		  - row-has-child-toggled (treepath, treeiter)
		  - row-deleted (treepath)

		Typically each signal should also flush the cache using
		C{self.cache.clear()}.

		@implementation: must be implemented by a subclass
		'''
		raise NotImplementedError

	def teardown(self):
		self.flush_cache()
		self.disconnect_all()

	def n_children_top(self):
		'''Return the number of items in the top level of the model'''
		raise NotImplementedError

	def get_mytreeiter(self, treepath):
		'''Returns a C{treeiter} object for C{treepath} or C{None}
		@implementation: must be implemented by a subclass
		'''
		raise NotImplementedError

	def find(self, obj):
		'''Return the treepath for a index object like a L{Path} or L{IndexTag}
		@raises IndexNotFoundError: if C{indexpath} is not found
		@implementation: must be implemented by a subclass
		'''
		raise NotImplementedError

	def find_all(self, obj):
		'''Like L{find()} but can return multiple results
		@implementation: must be implemented by subclasses that have mutiple
		entries for the same object. Default falls back to result of L{find()}.
		'''
		return [self.find(obj)]




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

from functools import partial

import logging

logger = logging.getLogger('zim.export')

from zim.base import MovingWindowIter

from zim.config import data_file
from zim.notebook import Path
from zim.formats import get_format

from zim.export.exporters import Exporter, createIndexPage
from zim.export.linker import ExportLinker
from zim.export.template import ExportTemplateContext

from zim.fs import adapt_from_oldfs
from zim.newfs import FileNotFoundError, LocalFolder



class FilesExporterBase(Exporter):
	'''Base class for exporters that export to files'''

	def __init__(self, layout, template, format, document_root_url=None):
		'''Constructor
		@param layout: a L{ExportLayout} to map pages to files
		@param template: a L{Template} object
		@param format: the format for the file content
		@param document_root_url: optional URL for the document root
		'''
		self.layout = layout
		self.template = template
		self.format = get_format(format) # XXX
		self.document_root_url = document_root_url

	def export_attachments_iter(self, notebook, page):
		# XXX FIXME remove need for notebook here
		# XXX what to do with folders that do not map to a page ?
		source = notebook.get_attachments_dir(page)
		target = self.layout.attachments_dir(page)
		try:
			for file in source.list_files():
					yield file
					targetfile = target.file(file.basename)
					if targetfile.exists():
						targetfile.remove() # Export does overwrite by default
					file.copyto(targetfile)
		except FileNotFoundError:
			pass

	def export_resources(self):
		dir = self.layout.resources_dir()

		# Copy icons -- TODO should we define default resources somewhere ?
		#~ for name in ('checked-box', 'unchecked-box', 'xchecked-box'):
			#~ icon = data_file('pixmaps/%s.png' % name)
			#~ file = dir.file(name + '.png')
			#~ icon.copyto(file)

		# Copy template resources (can overwrite icons)
		if self.template.resources_dir \
		and self.template.resources_dir.exists():
			if dir.exists(): # Export does overwrite by default
				dir.remove_children()
				dir.remove()

			resources = self.template.resources_dir
			resources = adapt_from_oldfs(resources)
			resources.copyto(dir)


class MultiFileExporter(FilesExporterBase):
	'''Exporter that exports each page to a single file'''

	def __init__(self, layout, template, format, index_page=None, document_root_url=None):
		'''Constructor
		@param layout: a L{ExportLayout} to map pages to files
		@param template: a L{Template} object
		@param format: the format for the file content
		@param index_page: a page to output the index or C{None}
		@param document_root_url: optional URL for the document root
		'''
		FilesExporterBase.__init__(self, layout, template, format, document_root_url)
		if index_page:
			if isinstance(index_page, str):
				self.index_page = Path(Path.makeValidPageName(index_page))
			else:
				self.index_page = index_page
		else:
			self.index_page = None
		# TODO make index_page generic special page in output selection

	def export_iter(self, pages):
		self.export_resources()

		for prev, page, next in MovingWindowIter(pages):
			yield page
			try:
				self.export_page(pages.notebook, page, pages, prevpage=prev, nextpage=next)
					# XXX FIXME remove need for notebook here
				for file in self.export_attachments_iter(pages.notebook, page):
					yield file
					# XXX FIXME remove need for notebook here
			except:
				logger.exception('Error while exporting: %s', page.name)
				raise

		if self.index_page:
			try:
				logger.info('Export index: %s', self.index_page)
				yield self.index_page
				self.export_index(self.index_page, pages)
			except:
				logger.exception('Error while exporting index')

	def export_page(self, notebook, page, pages, prevpage=None, nextpage=None):
		# XXX FIXME remove need for notebook here

		file = self.layout.page_file(page)
		if file.exists():
			file.remove() # export does overwrite by default

		linker_factory = partial(ExportLinker,
			notebook=notebook,
			layout=self.layout,
			output=file,
			usebase=self.format.info['usebase'],
			document_root_url=self.document_root_url
		)
		dumper_factory = self.format.Dumper # XXX

		context = ExportTemplateContext(
			notebook,
			linker_factory, dumper_factory,
			title=page.get_title(),
			content=[page],
			home=notebook.get_home_page(),
			up=None, # TODO
			prevpage=prevpage, nextpage=nextpage,
			links={'index': self.index_page},
			index_generator=pages.index,
			index_page=page,
		)

		lines = []
		self.template.process(lines, context)
		file.writelines(lines)

	def export_index(self, index_page, pages):
		if pages.prefix:
			index_page = pages.prefix + index_page

		page = createIndexPage(pages.notebook, index_page, pages.prefix)
		self.export_page(pages.notebook, page, pages)


class SingleFileExporter(FilesExporterBase):
	'''Exporter that exports all page to the same file'''

	# TODO make robust for errors during page iteration - needs to be in template code - allow removing try .. except in multifile above ?

	def export_iter(self, pages):
		self.export_resources()

		linker_factory = partial(ExportLinker,
			notebook=pages.notebook,
			layout=self.layout,
			output=self.layout.file,
			usebase=self.format.info['usebase'],
			document_root_url=self.document_root_url
		)
		dumper_factory = self.format.Dumper # XXX

		context = ExportTemplateContext(
			pages.notebook,
			linker_factory, dumper_factory,
			title=pages.title, # XXX
			content=pages,
			special=None, # TODO
			home=None,  # TODO
			links=None,
			index_generator=pages.index,
			index_page=None,
		)

		lines = []
		self.template.process(lines, context)
		if self.layout.file.exists():
			self.layout.file.remove() # export does overwrite by default
		self.layout.file.writelines(lines)

		# TODO incremental write to save memory on large notebooks...
		# TODO also yield while exporting main page

		for page in pages:
			yield page
			for file in self.export_attachments_iter(pages.notebook, page):
				yield file


#~ class StaticFileExporter(SingleFileExporter):

	# Single file, but link files with absolute path
	# used e.g. for print-to-browser

	# TODO overload to prevent copying files




# Copyright 2009-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GObject
from gi.repository import Pango

import logging
import re

import zim.datetimetz as datetime
from zim.base.naturalsort import natural_sorted

from zim.notebook import Path
from zim.actions import toggle_action, initialize_actiongroup, PRIMARY_MODIFIER_MASK
from zim.signals import DelayedCallback, SIGNAL_AFTER, SignalHandler, ConnectorMixin
from zim.config import ConfigManager, ConfigDefinition
from zim.config.dicts import String
from zim.plugins import PluginManager, ExtensionBase, extendable
from zim.gui.widgets import \
	Dialog, WindowSidePaneWidget, InputEntry, \
	BrowserTreeView, SingleClickTreeView, ScrolledWindow, HPaned, StatusPage, \
	encode_markup_text, decode_markup_text, widget_set_css, \
	uistate_property
from zim.gui.actionextension import ActionExtensionBase, populate_toolbar_with_actions
from zim.gui.clipboard import Clipboard


logger = logging.getLogger('zim.plugins.tasklist')

from .indexer import AllTasks, ActiveTasks, InboxTasks, NextActionTasks, OpenProjectsTasks, WaitingTasks, \
	_MAX_DUE_DATE, _MIN_START_DATE, _NO_TAGS, _date_re, _tag_re, _parse_task_labels, _task_labels_re, \
	TASK_STATUS_OPEN, TASK_STATUS_CLOSED, TASK_STATUS_CANCELLED, TASK_STATUS_MIGRATED, TASK_STATUS_TRANSMIGRATED


# Selection lists
SELECTION_ALL = 'all'
SELECTION_ACTIVE = 'active'
SELECTION_INBOX = 'inbox'
SELECTION_NEXT = 'next'
SELECTION_PROJECTS = 'projects'
SELECTION_WAITING = 'waiting'

# Model Columns
VIS_COL = 0 # visible
ACT_COL = 1 # actionable
PRIO_COL = 2
START_COL = 3
DUE_COL = 4
TAGS_COL = 5
DESC_COL = 6
PAGE_COL = 7
TASKID_COL = 8
PRIO_SORT_COL = 9
PRIO_SORT_LABEL_COL = 10
STATUS_COL = 11
STATUS_ICON_NAME_COL = 12


class TaskStatusUIState(ConfigDefinition):
	'''Data type for selected task status list in uistate'''

	def check(self, value):
		# Valid data is sequence of integers for the TASK_STATUS_* constants
		if isinstance(value, str):
			try:
				value = [int(s) for s in value.split(',')]
			except:
				raise ValueError
		else:
			if not (
				isinstance(value, (list, tuple))
				and all(isinstance(v, int) for v in value)
			):
				raise ValueError

		if not all((v >= 0 and v <= 4) for v in value):
			# In range of TASK_STATUS_OPEN to TASK_STATUS_TRANSMIGRATED
			raise ValueError

		return value


class TaskListWidgetMixin(object):
	'''Common functions between side-pane widget and dialog'''

	SELECTION_MAP = { # Define in class to make sure translations are intialized
		SELECTION_ALL: (_('All Tasks'), AllTasks), # T: list with tasks
		SELECTION_ACTIVE: (_('Active'), ActiveTasks), # T: list with tasks
		SELECTION_INBOX: (' \u2012 ' + _('Inbox'), InboxTasks), # T: list with tasks (GTD style category)
		SELECTION_NEXT: (' \u2012 ' + _('Next Actions'), NextActionTasks), # T: list with tasks (GTD style category)
		SELECTION_PROJECTS: (' \u2012 ' + _('Projects'), OpenProjectsTasks), # T: list with tasks (GTD style category)
		SELECTION_WAITING: (_('Waiting'), WaitingTasks), # T: list with tasks (GTD style category)
	}

	taskselection_type = uistate_property('task_list', SELECTION_ALL, tuple(SELECTION_MAP.keys()))
	status = uistate_property('task_status', [TASK_STATUS_OPEN], TaskStatusUIState)

	def __init__(self, index, uistate, properties):
		self.index = index
		self.tasklisttreeview = None
		self.selection_list = None
		self.tag_list = None
		self._mbutton = None

		self.uistate = uistate
		self.uistate.setdefault('sort_column', 0)
		self.uistate.setdefault('sort_order', int(Gtk.SortType.DESCENDING))

		self.taskselection = self.SELECTION_MAP[self.taskselection_type][1].new_from_index(index)
		self.label_tag_filter = (None, None, None) # NOTE: Not taken from uistate since encoding&validation would be non-trivial

		self.connectto(properties, 'changed', self.on_properties_changed)

	def _init_selection_state(self):
		# Ensure widgets reflect selection
		state = self._get_selection_state()
		self._set_selection_state(state)

	def _get_selection_state(self):
		return (self.taskselection_type, self.status[:], self.label_tag_filter)

	def _set_selection_state(self, state):
		self.status = state[1]
		self.selection_list._select(state[0])
		self.tag_list._set_selected_labels_tags_pages(*state[2])

	def _create_tasklisttreeview(self, opener, properties):
		self.tasklisttreeview = TaskListTreeView(
			self.taskselection, opener,
			_parse_task_labels(properties['labels']),
			use_workweek=properties['use_workweek'],
			sort_column=self.uistate['sort_column'],
			sort_order=self.uistate['sort_order'],
		)
		stack = Gtk.Stack()
		for name, widget in (
			('treeview', ScrolledWindow(self.tasklisttreeview, shadow=Gtk.ShadowType.NONE)),
			('placeholder', StatusPage('task-list-closed-symbolic', _('No tasks'))), # T: placeholder label for sidepane
		):
			widget.show_all()
			stack.add_named(widget, name)

		def switch_placeholder(treeview, not_empty):
			if not_empty:
				stack.set_visible_child_name('treeview')
			else:
				stack.set_visible_child_name('placeholder')

		self.tasklisttreeview.connect('view-changed', switch_placeholder)
		stack.set_visible_child_name('treeview')

		return stack

	def _create_selection_menubutton(self, opener, properties, show_inbox_next):
		self._mbutton = Gtk.MenuButton()
		self._set_mbutton_label(self.SELECTION_MAP[SELECTION_ALL][0])
		popover = Gtk.Popover()
		self._mbutton.set_popover(popover)
		popover.add(self._create_selection_pane(properties, show_inbox_next))

		popout = Gtk.Button()
		icon = Gtk.Image.new_from_icon_name('window-pop-out-symbolic', Gtk.IconSize.SMALL_TOOLBAR)
		popout.add(icon)
		popout.set_tooltip_text(_('Show tasklist window')) # T: tooltip
		popout.set_alignment(0.5, 0.5)
		popout.set_relief(Gtk.ReliefStyle.NONE)
		popout.connect('clicked', lambda b: self._show_dialog_action(self._get_selection_state()))

		hbox = Gtk.HBox()
		hbox.set_border_width(3)
		hbox.set_spacing(5)
		hbox.pack_start(self._mbutton, True, True, 0)
		hbox.pack_start(popout, False, True, 0)

		return hbox

	def _create_selection_pane(self, properties, show_inbox_next, width=300):
		vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		vbox1 = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		vbox1.set_border_width(5) # Else scrollbar overlays numbers

		self.selection_list = ListSelectionView(show_inbox_next=show_inbox_next)
		self.selection_list.connect('row-activated', self.on_selection_activated)
		vbox1.add(self.selection_list)

		self.tag_list = LabelAndTagView(self.taskselection, self, _parse_task_labels(properties['labels']), properties['show_pages'])
		vbox1.add(self.tag_list)

		swindow = ScrolledWindow(vbox1, shadow=Gtk.ShadowType.NONE)
		swindow.set_size_request(width, 500)
		vbox.pack_start(swindow, True, True, 0)
		vbox.pack_end(self._create_clear_tags_button(), False, False, 0)
		vbox.show_all()
		return vbox

	def _create_clear_tags_button(self):
		button = Gtk.Button('Clear selection')
		button.connect('clicked', lambda o: self.tag_list.unselect_all())

		def update_button(tag_list):
			selected = tag_list.get_selected_rows()
			button.set_sensitive(bool(selected))

		update_button(self.tag_list)
		self.tag_list.connect('selected-rows-changed', update_button)
		return button

	def _create_filter_entry(self):
		filter_entry = InputEntry(placeholder_text=_('Filter tasks')) # T: label for filtering/searching tasks
		filter_entry.set_icon_to_clear()
		filter_cb = DelayedCallback(500,
			lambda o: self.on_filter_changed(filter_entry.get_text()))
		filter_entry.connect('changed', filter_cb)
		return filter_entry

	def on_filter_changed(self, text):
		self.tasklisttreeview.set_filter(text)

	def on_properties_changed(self, properties):
		task_labels = _parse_task_labels(properties['labels'])
		self.tasklisttreeview.update_properties(
			task_labels=task_labels,
			use_workweek=properties['use_workweek'],
		)
		self.tag_list.update_properties(
			task_labels=_parse_task_labels(properties['labels']),
			show_pages=properties['show_pages']
		)

	def reload_view(self):
		state = self._get_selection_state()
		for view in (self.tasklisttreeview, self.tag_list):
			if view is not None:
				view.refresh()
		self._set_selection_state(state)

	def on_selection_activated(self, listbox, boxrow):
		label = boxrow.get_children()[0]
		self.set_selection(label._zim_key)

	def set_selection(self, key):
		self.taskselection_type = key
		label, cls = self.SELECTION_MAP[key]
		self._set_mbutton_label(label)
		self.taskselection = cls.new_from_index(self.index)
		self.taskselection.set_status_included(*self.status)
		self.tasklisttreeview.set_taskselection(self.taskselection)
		self.tag_list.set_taskselection(self.taskselection)

	def _set_mbutton_label(self, text):
		if self._mbutton is None:
			return

		child = self._mbutton.get_children()[0]
		self._mbutton.remove(child)

		hbox = Gtk.HBox()
		label = Gtk.Label(text)
		label.set_alignment(0.1, 0.5)
		hbox.add(label)
		hbox.pack_end(Gtk.Image.new_from_icon_name('pan-down-symbolic', Gtk.IconSize.MENU), False, False, 0)
		hbox.show_all()
		self._mbutton.add(hbox)

	def set_label_tag_filter(self, labels, tags, pages):
		self.label_tag_filter = (labels, tags, pages)
		self.tasklisttreeview.set_label_tag_filter(labels, tags, pages)


class TaskListWidget(Gtk.VBox, TaskListWidgetMixin, WindowSidePaneWidget):

	title = _('Tas_ks') # T: tab label for side pane

	def __init__(self, index, opener, properties, show_inbox_next, uistate, show_dialog_action):
		GObject.GObject.__init__(self)
		TaskListWidgetMixin.__init__(self, index, uistate, properties)
		self._close_button = None
		self._show_dialog_action = show_dialog_action

		swindow = self._create_tasklisttreeview(opener, properties)
		self._header_hbox = self._create_selection_menubutton(opener, properties, show_inbox_next)
		self.pack_start(self._header_hbox, False, True, 0)
		self.pack_start(swindow, True, True, 0)

		self.tasklisttreeview.set_view_column_visible('status_icon', False) # no status toggle in this view
		self.tasklisttreeview.view_columns['task'].set_min_width(200) # don't let this column get too small

		filter_entry = self._create_filter_entry()
		self.pack_end(filter_entry, False, True, 0)

		self._init_selection_state()

	def set_embeded_closebutton(self, button):
		if self._close_button:
			self._header_hbox.remove(self._close_button)

		if button is not None:
			self._header_hbox.pack_end(button, False, True, 0)

		self._close_button = button
		return True


class TaskListWindowExtension(ActionExtensionBase):
	'''Base class for window extensions
	Actions can be defined using e.g. C{@action} decorator, see
	L{zim.actions}, and will automatically be added to the window on
	initialization.
	'''

	# TODO - should be generic base class for extending windows
	# TODO - remove actions on teardown !!!

	def __init__(self, plugin, window):
		ExtensionBase.__init__(self, plugin, window)
		self.window = window
		self.connectto(window, 'destroy')
		self._add_headerbar_actions()

	def on_destroy(self, dialog):
		self.destroy()


from zim.config import value_is_coord


@extendable(TaskListWindowExtension)
class TaskListWindow(TaskListWidgetMixin, ConnectorMixin, Gtk.Window):

	def __init__(self, notebook, index, navigation, properties, show_inbox_next, hide_on_close=True):
		Gtk.Window.__init__(self)
		self.uistate = notebook.state[self.__class__.__name__]
		defaultwindowsize=(550, 400)

		# Catching this signal prevents the window to actually be destroyed
		# when the user tries to close it. The action for close should either
		# hide or destroy the window.
		if hide_on_close:
			def do_delete_event(self, *a):
				logger.debug('Action: close tasklist window (delete-event)')
				self.save_uistate()
				self.hide()
				return True # Do not destroy - let close() handle it

			self.connect('delete-event', do_delete_event)
		else:
			self.connect('delete-event', self.save_uistate)

		# note: _windowpos is defined with a leading "_" so it is not
		# persistent across instances, this is intentional to avoid
		# e.g. messy placement for seldom used dialogs
		self.uistate.setdefault('_windowpos', None, check=value_is_coord)
		if self.uistate['_windowpos'] is not None:
			x, y = self.uistate['_windowpos']
			self.move(x, y)

		self.uistate.setdefault('windowsize', defaultwindowsize, check=value_is_coord)
		if self.uistate['windowsize'] is not None:
			w, h = self.uistate['windowsize']
			self.set_default_size(w, h)

		TaskListWidgetMixin.__init__(self, index, self.uistate, properties)

		vbox = Gtk.VBox()
		self.add(vbox)

		if_prefs = ConfigManager.preferences['GtkInterface']
		# Normally would define the 'show_headerbar' pref here, but don't want
		# to overrule the definition in mainwindow. Therefore just using `get()`
		if if_prefs.get('show_headerbar', True):
			self._headerbar = Gtk.HeaderBar()
			self._headerbar.set_title(_('Tasks') + '  -  ' + notebook.name) # T: Window title for tasklist
			self._headerbar.set_show_close_button(True)
			self.set_titlebar(self._headerbar)
			self._toolbar = None
		else:
			self.set_title(_('Tasks') + '  -  ' + notebook.name) # T: Window title for tasklist
			self._headerbar = None
			self._toolbar = Gtk.Toolbar()
			vbox.pack_start(self._toolbar, False, False, 0)

			def on_extensions_changed(o, obj):
				if obj is self:
					self._update_toolbar()

			self.connectto(PluginManager, 'extensions-changed', on_extensions_changed)

		self.hpane = HPaned()
		self.uistate.setdefault('hpane_pos', 75)
		self.hpane.set_position(self.uistate['hpane_pos'])
		pane2_vbox = Gtk.VBox()
		self.hpane.pack2(pane2_vbox, resize=True)

		vbox.add(self.hpane)

		self._search_bar = Gtk.SearchBar()
		pane2_vbox.pack_start(self._search_bar, False, True, 0)
		self._filter_entry = self._create_filter_entry()
		self._filter_entry.set_size_request(400, -1) # FIXME, scale as percentage of window size / pane size ?
		self._search_bar.add(self._filter_entry)
		self._search_bar.connect_entry(self._filter_entry)
		self._filter_entry.show_all()
		self.connect('key-press-event', lambda o, e: self._search_bar.handle_event(e))
		self._search_bar.connect('notify::search-mode-enabled', lambda *a: self.show_search(active=self._search_bar.get_search_mode()))
			# make toggle button consistent when e.g. closing bar with <Escape> key binding

		swindow = self._create_tasklisttreeview(navigation, properties)
		pane2_vbox.pack_start(swindow, True, True, 0)

		self.tasklisttreeview.view_columns['task'].set_min_width(400) # don't let this column get too small

		self.hpane.pack1(self._create_selection_pane(properties, show_inbox_next, width=150), resize=False)

		self._init_count_label()
		self.tasklisttreeview.connect('view-changed', self._update_count_label)

		if self._headerbar:
			popover = self._create_menu()
			button = self._create_menu_button(popover)
			self._headerbar.pack_end(button)
			self._headerbar.pack_end(self.show_search.create_icon_button())
			self._headerbar.pack_end(self._count_label)
		else:
			assert self._toolbar is not None
			self._update_toolbar()

		self._init_selection_state()

		# Initialze actions
		initialize_actiongroup(self, 'win')

		# Hack to add Ctrl-F keybinding for the filter
		# FIXME: would prefer Window base class to handle actions out of the box
		group = Gtk.AccelGroup()
		group.connect( # <Primary><F>
			Gdk.unicode_to_keyval(ord('f')), PRIMARY_MODIFIER_MASK, Gtk.AccelFlags.VISIBLE,
			lambda *a: self.show_search() )
		self.add_accel_group(group)

	def _init_count_label(self):
		self._count_label = Gtk.Label()
		self._count_label.set_margin_start(12)
		self._count_label.set_margin_end(12)
		#context = self._count_label.get_style_context()
		#context.add_class(Gtk.STYLE_CLASS_SUBTITLE)
		model = self.tasklisttreeview.get_model()
		count = len(model) if model else 0
		self._update_count_label(None, count)

	def _update_count_label(self, view, count):
		self._count_label.set_text(_('%s shown') % str(count)) # T: number of tasks shown in tasklist window

	def _update_toolbar(self):
		for item in self._toolbar.get_children():
			self._toolbar.remove(item)

		populate_toolbar_with_actions(self._toolbar, self, include_headercontrols=True)

		space = Gtk.SeparatorToolItem()
		space.set_draw(False)
		space.set_expand(True)
		self._toolbar.insert(space, -1)

		item = Gtk.ToolItem()
		item.add(self._count_label)
		self._toolbar.insert(item, -1)

		item = self.show_search.create_tool_button(connect_button=False)
		item.set_action_name('win.show_search')
		self._toolbar.insert(item, -1)

		popover = self._create_menu()
		button = self._create_menu_toolbutton(popover)
		self._toolbar.insert(button, -1)

		self._toolbar.show_all()

	def _create_menu_button(self, popover):
		button = Gtk.MenuButton()
		button.set_direction(Gtk.ArrowType.NONE)
		button.set_popover(popover)
		button.show_all()
		return button

	def _create_menu_toolbutton(self, popover):
		button = Gtk.ToggleToolButton()
		button.set_icon_name('open-menu-symbolic')
		popover.set_relative_to(button)
		def toggle_popover(button):
			if button.get_active():
				popover.popup()
			else:
				popover.popdown()
		button.connect('toggled', toggle_popover)
		popover.connect('closed', lambda o: button.set_active(False))
		return button

	def _create_menu(self):
		popover = Gtk.Popover()

		vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		for key, label in (
			(TASK_STATUS_OPEN, _('Show open tasks')), # T: selection menu option
			(TASK_STATUS_CLOSED, _('Show closed tasks')), # T: selection menu option
			(TASK_STATUS_CANCELLED, _('Show cancelled tasks')), # T: selection menu option
			(TASK_STATUS_MIGRATED, _('Show migrated tasks')), # T: selection menu option
			(TASK_STATUS_TRANSMIGRATED, _('Show trans-migrated tasks')) # T: selection menu option
		):
			checkbutton = Gtk.CheckButton.new_with_label(label)
			checkbutton.set_active(key in self.status)
			checkbutton.connect('toggled', self.on_status_checkbox_toggled, key)
			vbox.add(checkbutton)

		popover.add(vbox)
		vbox.show_all()
		return popover

	def on_status_checkbox_toggled(self, checkbox, key):
		self.status = [s for s in self.status if s != key]
		if checkbox.get_active():
			self.status.append(key)
		self.taskselection.set_status_included(*self.status)
		self.tasklisttreeview.refresh()
		self.tag_list.refresh()

	def present(self):
		Gtk.Window.present(self)
		self.reload_view() # Might be out of date when we are hidden

	def reload_view(self):
		# Overloaded to only run what visible, force refresh on present() above
		state = self._get_selection_state()
		for view in (self.tasklisttreeview, self.tag_list):
			if view is not None and view.is_visible():
				view.refresh()
		self._set_selection_state(state)

	def save_uistate(self, *a):
		self.uistate['hpane_pos'] = self.hpane.get_position()

		for column in self.tasklisttreeview.get_columns():
			if column.get_sort_indicator():
				self.uistate['sort_column'] = column.get_sort_column_id()
				self.uistate['sort_order'] = int(column.get_sort_order())
				break
		else:
			# if it is unsorted, just use the defaults
			self.uistate['sort_column'] = PRIO_COL
			self.uistate['sort_order'] = Gtk.SortType.ASCENDING

		try:
			x, y = self.get_position()
			self.uistate['_windowpos'] = (x, y)
			w, h = self.get_size()
			self.uistate['windowsize'] = (w, h)
		except:
			logger.exception('Exception in save_uistate')

	@toggle_action(_('_Search'), '<Primary>F', verb_icon='edit-find-symbolic') # T: Menu item
	def show_search(self, show):
		self._search_bar.set_search_mode(show)



class ListSelectionView(Gtk.ListBox):

	def __init__(self, show_inbox_next=False):
		Gtk.ListBox.__init__(self)
		if show_inbox_next:
			lists = (
				SELECTION_ALL,
				SELECTION_ACTIVE,
				SELECTION_INBOX,
				SELECTION_NEXT,
				SELECTION_PROJECTS,
				SELECTION_WAITING,
			)
		else:
			lists = (
				SELECTION_ALL,
				SELECTION_ACTIVE,
				SELECTION_WAITING,
			)

		for key in lists:
			mylabel = Gtk.Label(TaskListWidgetMixin.SELECTION_MAP[key][0])
			mylabel.set_alignment(0.0, 0.5)
			mylabel._zim_key = key
			self.insert(mylabel, -1)

		widget_set_css(self, self.__class__.__name__, 'background-color: rgba(0.0, 0.0, 0.0, 0.0)')
			# Removing white background for listbox

		self.set_header_func(self._update_header_func)

	def _update_header_func(self, row, before):
		# Only set header once on first row
		if before is None and not row.get_header():
			label = Gtk.Label()
			label.set_markup('<b>%s</b>' % _('Lists')) # T: Header for list selection in tasklist window
			row.set_header(label)

	def _select(self, key):
		for row in self.get_children():
			if row.get_child()._zim_key == key:
				self.select_row(row)
				self.emit('row-activated', row)
				return True
		else:
			# NOTE: do not make this a fatal error, might be old uistate in place
			logger.debug("Could not find selection key: %s" % key)
			return False


class LabelAndTagView(Gtk.ListBox):

	def __init__(self, taskselection, tasklist_widget, task_labels, show_pages=True):
		Gtk.ListBox.__init__(self)

		self.taskselection = taskselection
		self.tasklist_widget = tasklist_widget
		self.task_labels = task_labels
		self.show_pages = show_pages

		widget_set_css(self, self.__class__.__name__, 'background-color: rgba(0.0, 0.0, 0.0, 0.0)')
			# Removing white background for listbox

		self.set_selection_mode(Gtk.SelectionMode.MULTIPLE)

		self.set_header_func(self._update_header_func)
		self.set_filter_func(self._filter_func)
		self.refresh()

	def update_properties(self, show_pages, task_labels):
		self.show_pages = show_pages
		self.task_labels = task_labels
		self.refresh()

	def do_button_release_event(self, event):
		# Implement behavior for de-selecting rows
		if event.type == Gdk.EventType.BUTTON_RELEASE \
		and event.button == 1 and not event.get_state() & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.META_MASK):
			x, y = list(map(int, event.get_coords()))
			row = self.get_row_at_y(y)
			if row and row.is_selected():
				self.unselect_row(row)
				return True

		return Gtk.ListBox.do_button_release_event(self, event)

	def do_selected_rows_changed(self):
		labels, tags, pages = self._get_selected_labels_tags_pages()
		self.tasklist_widget.set_label_tag_filter(labels, tags, pages)
		self.update()

	def _get_selected_labels_tags_pages(self):
		labels, tags, pages = [], [], []
		for row in self.get_selected_rows():
			if row._zim_type == 'label':
				labels.append(row._zim_label)
			elif row._zim_type == 'tag':
				tags.append(row._zim_label)
			else: # 'page'
				pages.append(row._zim_label)
		return labels, tags, pages

	def _set_selected_labels_tags_pages(self, labels, tags, pages):
		selection = {'label': labels, 'tag': tags, 'page': pages}
		for row in self.get_children():
			myselection = selection[row._zim_type]
			if myselection is not None and row._zim_label in myselection:
					self.select_row(row)
			else:
				self.unselect_row(row)

	def set_taskselection(self, taskselection):
		self.taskselection = taskselection
		self.refresh()

	def refresh(self):
		for child in self.get_children():
			self.remove(child)

		labels, tags, pages = self.taskselection.count_labels_and_tags_pages(self.task_labels)

		for label in self.task_labels: # Keep original order
			count = labels.get(label, 0)
			self.add(self._create_item(label, count, 'label'))

		for tag in natural_sorted(tags.keys()):
			count = tags[tag]
			if tag == _NO_TAGS:
				row = self._create_item(_('Untagged'), count, 'tag') # T: selection of tasks without tags
				row._zim_label = _NO_TAGS
				self.add(row)
			else:
				self.add(self._create_item(tag, count, 'tag'))

		if self.show_pages:
			for name in natural_sorted(pages.keys()):
				count = pages[name]
				self.add(self._create_item(name, count, 'page'))

		self.show_all()

	def update(self):
		intersect = self._get_selected_labels_tags_pages()
		labels, tags, pages = self.taskselection.count_labels_and_tags_pages(self.task_labels, intersect)
		tags = {k.lower(): c for k, c in tags.items()} # Convert to lower because potential case mismatch with labels
		for row in self.get_children():
			if row._zim_type == 'label':
				count = labels.get(row._zim_label, 0)
			elif row._zim_type == 'tag':
				count = tags.get(row._zim_label.lower(), 0)
			else: #'page'
				count = pages.get(row._zim_label, 0)
			hbox = row.get_child()
			count_label = hbox.get_children()[-1]
			count_label.set_text(str(count))
			row._zim_count = count
			row.changed()

	def _create_item(self, label, count, type):
		hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
		glabel = Gtk.Label(label)
		glabel.set_alignment(0.0, 0.5)
		glabel.set_ellipsize(Pango.EllipsizeMode.END)
		countlabel = Gtk.Label(str(count))
		countlabel.set_sensitive(False) # Lighter color compared to main label
		hbox.pack_start(glabel, True, True, 0)
		hbox.pack_start(countlabel, False, False, 0)

		row = Gtk.ListBoxRow()
		row.add(hbox)
		row._zim_type = type
		row._zim_label = label
		row._zim_count = count
		row.set_activatable(False)
		return row

	def _filter_func(self, row):
		return row._zim_count > 0

	def _update_header_func(self, row, before):
		if before is None or before._zim_type != row._zim_type:
			if not row.get_header():
				if row._zim_type == 'label':
					text = _('Labels') # T: header in selection drop down
				elif row._zim_type == 'tag':
					text = _('Tags') # T: header in selection drop down
				else: # 'page'
					text = _('Page') # T: header in selection drop down
				label = Gtk.Label()
				label.set_markup('<b>%s</b>' % text)
				row.set_header(label)
		else:
			if row.get_header():
				row.set_header(None)


_cal_days_to_work_days = [
	# for each weekday 5 offsets used in algo below
	# represent weekends in a 14 day period starting at given weekday
	# first number is offset for weekend at start of the range
	# next 4 numbers are weekend days in the range
	None,
	[0, 5, 6, 12, 13], # monday
	[0, 4, 5, 11, 12], # tuesday
	[0, 3, 4, 10, 11], # wednesday
	[0, 2, 3,  9, 10], # thursday
	[0, 1, 2,  8,  9], # friday
	[1, 5, 6, 12, 13], # saturday
	[0, 6, 7, 13, 14], # sunday
]


def days_to_str(days, use_workweek, weekday):
	# days are calendar days, not working days
	# convert to working days if period is less than 2 weeks
	if days >= 300:
		return '%iy' % round(float(days) / 365) # round up to 1 year from ~10 months
	elif days >= 28:
		return '%im' % round(float(days) / 30) # round up to 1 year from 4 calendar weeks
	elif days >= 14:
		return '%iw' % round(float(days) / 7)
	elif use_workweek:
		offsets = _cal_days_to_work_days[weekday]
		days -= offsets[0]
		if days >= offsets[4]:
			days -= 4
		elif days == offsets[3]:
			days -= 3
		elif days >= offsets[2]:
			days -= 2
		elif days == offsets[1]:
			days -= 1
		return '%id' % days
	else:
		return '%id' % days


class TaskListTreeView(BrowserTreeView):

	# These default values are overwritten based on "styles.conf" configuration
	HIGH_COLOR = '#ef2929' # red (derived from Tango style guide)
	MEDIUM_COLOR = '#f57900' # orange
	ALERT_COLOR = '#fce947' # yellow

	PRIO_COLORS = [None, ALERT_COLOR, MEDIUM_COLOR, HIGH_COLOR] # index 0..3

	TAG_TEXT_COLOR = '#ce5c00'
	INACTIVE_TEXT_COLOR = 'darkgrey'

	# define signals we want to use - (closure type, return type and arg types)
	__gsignals__ = {
		'view-changed': (GObject.SignalFlags.RUN_LAST, None, (object,)),
	}

	def __init__(self,
		taskselection, opener,
		task_labels,
		use_workweek=False,
		sort_column=PRIO_COL, sort_order=Gtk.SortType.DESCENDING
	):
		self.text_style = ConfigManager.get_config_dict('style.conf')
		self.text_style['TaskList Prio High'].define({'background': String(self.HIGH_COLOR)})
		self.text_style['TaskList Prio Medium'].define({'background': String(self.MEDIUM_COLOR)})
		self.text_style['TaskList Prio Alert'].define({'background': String(self.ALERT_COLOR)})
		self.text_style['TaskList Inactive'].define({'foreground': String(self.INACTIVE_TEXT_COLOR)})
		self.text_style.connect('changed', lambda o: self.on_text_style_changed())
		self.on_text_style_changed()

		self.real_model = Gtk.TreeStore(bool, bool, int, str, str, object, str, str, int, int, str, int, str)
			# VIS_COL, ACT_COL, PRIO_COL, START_COL, DUE_COL, TAGS_COL, DESC_COL, PAGE_COL, TASKID_COL, PRIO_SORT_COL, PRIO_SORT_LABEL_COL, STATUS_COL, STATUS_ICON_NAME_COL
		model = self.real_model.filter_new()
		model.set_visible_column(VIS_COL)
		model = Gtk.TreeModelSort(model)
		model.set_sort_column_id(sort_column, sort_order)
		BrowserTreeView.__init__(self, model)
		self.set_headers_visible(True)

		self.taskselection = taskselection
		self.opener = opener
		self.filter = None
		self.tag_filter = None
		self.label_filter = None
		self.page_filter = None
		self.task_labels = task_labels
		self.use_workweek = use_workweek
		self._render_waiting_actionable = False
		self.view_columns = {}
		self._visible_columns = {}

		self._icon_names = {
			TASK_STATUS_OPEN: 'task-list-open-symbolic',
			TASK_STATUS_CLOSED: 'task-list-closed-symbolic',
			TASK_STATUS_CANCELLED: 'task-list-cancelled-symbolic',
			TASK_STATUS_MIGRATED: 'task-list-migrated-symbolic',
			TASK_STATUS_TRANSMIGRATED: 'task-list-transmigrated-symbolic',
		}

		# Status column
		cell_renderer = Gtk.CellRendererPixbuf()
		column = Gtk.TreeViewColumn(' ', cell_renderer)
		column.add_attribute(cell_renderer, 'icon-name', STATUS_ICON_NAME_COL)
		column.set_sort_column_id(STATUS_COL)
		self.append_column(column)
		self.view_columns['status_icon'] = column

		# Add some rendering for the Prio column
		def render_prio(col, cell, model, i, data):
			text = model.get_value(i, PRIO_SORT_LABEL_COL)
			if not model.get_value(i, ACT_COL):
				text = '<span color="%s">%s</span>' % (self.INACTIVE_TEXT_COLOR, text)
				bg = None
			else:
				prio = model.get_value(i, PRIO_COL)
				bg = self.PRIO_COLORS[min(prio, 3)]
			cell.set_property('markup', text)
			cell.set_property('cell-background', bg)

		cell_renderer = Gtk.CellRendererText()
		column = Gtk.TreeViewColumn('!', cell_renderer)
		column.set_cell_data_func(cell_renderer, render_prio)
		column.set_sort_column_id(PRIO_SORT_COL)
		self.append_column(column)
		self.view_columns['prio'] = column

		# Rendering for task description column
		cell_renderer = Gtk.CellRendererText()
		cell_renderer.set_property('ellipsize', Pango.EllipsizeMode.END)
		column = Gtk.TreeViewColumn(_('Task'), cell_renderer, markup=DESC_COL)
				# T: Column header Task List dialog
		column.set_resizable(True)
		column.set_sort_column_id(DESC_COL)
		column.set_expand(True)
		self.append_column(column)
		self.set_expander_column(column)
		self.view_columns['task'] = column

		# custom tooltip
		self.props.has_tooltip = True
		self.connect("query-tooltip", self._query_tooltip_cb)

		# Rendering of the Date column
		day_of_week = datetime.date.today().isoweekday()
		if use_workweek and day_of_week == 4:
			# Today is Thursday - 2nd day ahead is after the weekend
			delta1, delta2 = 1, 3
		elif use_workweek and day_of_week == 5:
			# Today is Friday - next day ahead is after the weekend
			delta1, delta2 = 3, 4
		else:
			delta1, delta2 = 1, 2

		today = str(datetime.date.today())
		tomorrow = str(datetime.date.today() + datetime.timedelta(days=delta1))
		dayafter = str(datetime.date.today() + datetime.timedelta(days=delta2))
		def render_date(col, cell, model, i, model_col):
			date = model.get_value(i, model_col)
			if date in (_MAX_DUE_DATE, _MIN_START_DATE):
				cell.set_property('text', '')
			else:
				if not model.get_value(i, ACT_COL):
					date = '<span color="%s">%s</span>' % (self.INACTIVE_TEXT_COLOR, date)
				cell.set_property('markup', date)
				# TODO allow strftime here

			if model_col == DUE_COL:
				if date <= today:
					color = self.HIGH_COLOR
				elif date <= tomorrow:
					color = self.MEDIUM_COLOR
				elif date <= dayafter:
					# "<=" because tomorrow and/or dayafter can be after the weekend
					color = self.ALERT_COLOR
				else:
					color = None
				cell.set_property('cell-background', color)


		for key, col, label in (
			('due', DUE_COL, _('Due')), # T: Column header Task List dialog
			('start', START_COL, _('Start')), # T: Column header Task List dialog
		):
			cell_renderer = Gtk.CellRendererText()
			column = Gtk.TreeViewColumn(label, cell_renderer)
			column.set_cell_data_func(cell_renderer, render_date, col)
			column.set_sort_column_id(col)
			self.append_column(column)
			self.view_columns[key] = column

		# Rendering for page name column
		def render_page(col, cell, model, i, data):
			text = model.get_value(i, PAGE_COL)
			text = encode_markup_text(text)
			if not model.get_value(i, ACT_COL):
				text = '<span color="%s">%s</span>' % (self.INACTIVE_TEXT_COLOR, text)
			cell.set_property('markup', text)

		cell_renderer = Gtk.CellRendererText()
		column = Gtk.TreeViewColumn(_('Page'), cell_renderer)
				# T: Column header Task List dialog
		column.set_cell_data_func(cell_renderer, render_page)
		column.set_sort_column_id(PAGE_COL)
		self.append_column(column)
		self.view_columns['page'] = column

		# Finalize
		for key in self.view_columns:
			self._visible_columns[key] = True # Default all visible
		self.refresh()

		# HACK because we can not register ourselves :S
		self.connect('row_activated', self.__class__.do_row_activated)
		self.connect('focus-in-event', self.__class__.do_focus_in_event)

	def on_text_style_changed(self, *a):
		self.HIGH_COLOR = self.text_style['TaskList Prio High']['background']
		self.MEDIUM_COLOR = self.text_style['TaskList Prio Medium']['background']
		self.ALERT_COLOR = self.text_style['TaskList Prio Alert']['background']
		self.PRIO_COLORS = [None, self.ALERT_COLOR, self.MEDIUM_COLOR, self.HIGH_COLOR] # index 0..3

		self.TAG_TEXT_COLOR = self.text_style['Tag tag'].get('foreground', self.TAG_TEXT_COLOR)
		self.INACTIVE_TEXT_COLOR = self.text_style['TaskList Inactive']['foreground']

	def get_view_column_visible(self, key):
		assert key in self.view_columns
		return self.view_columns[key].get_visible()

	def set_view_column_visible(self, key, visible):
		assert key in self.view_columns
		self._visible_columns[key] = visible
		if self.taskselection.STYLE == 'inbox' and key in ('prio', 'due'):
			pass
		else:
			self.view_columns[key].set_visible(visible)

	def set_taskselection(self, taskselection):
		self.taskselection = taskselection

		if self.taskselection.STYLE == 'inbox':
			for key in ('prio', 'due'):
				self.view_columns[key].set_visible(False)
		else:
			for key in ('prio', 'due'):
				self.view_columns[key].set_visible(self._visible_columns[key])

		if self.taskselection.STYLE == 'waiting':
			self._render_waiting_actionable = True
		else:
			self._render_waiting_actionable = False

		self.refresh()

	def update_properties(self,
		task_labels=None,
		use_workweek=None,
	):
		if task_labels is not None:
			self.task_labels = task_labels

		if use_workweek is not None:
			self.use_workweek = use_workweek

		self.refresh()

	def refresh(self):
		'''Refresh the model based on index data'''
		self.real_model.clear() # flush
		self._append_tasks(self.taskselection.list_tasks(), None)

		self._today = datetime.date.today()
		self._eval_filter() # keep current selection
		self.expand_all()

	def _append_tasks(self, task_iter, parent_tree_iter):
		task_label_re = _task_labels_re(self.task_labels)
		today = datetime.date.today()
		today_str = str(today)
		weekday = today.isoweekday()

		for prio_sort_int, row in enumerate(task_iter):
			path = Path(row['name'])
			tags = [t for t in row['tags'].split(',') if t]
			lowertags = [t.lower() for t in tags]
			actionable = self._render_waiting_actionable or not row['waiting']

			# Checkbox
			status = row['status']
			status_icon_name = self._icon_names[status]

			# Format label for "prio" column
			if status != TASK_STATUS_OPEN:
				prio_sort_label = '!' * min(row['prio'], 3)
			elif row['start'] > today_str:
				actionable = False
				y, m, d = row['start'].split('-')
				td = datetime.date(int(y), int(m), int(d)) - today
				prio_sort_label = '>' + days_to_str(td.days, self.use_workweek, weekday)
				if row['prio'] > 0:
					prio_sort_label += ' ' + '!' * min(row['prio'], 3)
			elif row['due'] < _MAX_DUE_DATE:
				y, m, d = row['due'].split('-')
				td = datetime.date(int(y), int(m), int(d)) - today
				prio_sort_label = \
					'!' * min(row['prio'], 3) + ' ' if row['prio'] > 0 else ''
				if td.days < 0:
						prio_sort_label += '<b><u>OD</u></b>' # over due
				elif td.days == 0:
						prio_sort_label += '<u>TD</u>' # today
				else:
						prio_sort_label += days_to_str(td.days, self.use_workweek, weekday)
			else:
				prio_sort_label = '!' * min(row['prio'], 3)

			# Format description
			desc = _date_re.sub('', row['description'])
			desc = re.sub(r'\s*!+\s*', ' ', desc) # get rid of exclamation marks
			desc = encode_markup_text(desc)
			if actionable:
				desc = _tag_re.sub(r'<span color="%s">@\1</span>' % self.TAG_TEXT_COLOR, desc) # highlight tags
				desc = task_label_re.sub(r'<b>\1</b>', desc) # highlight labels
			else:
				desc = '<span color="%s">%s</span>' % (self.INACTIVE_TEXT_COLOR, desc)

			# Insert all columns
			modelrow = [False, actionable, row['prio'], row['start'], row['due'], tags, desc, path.name, row['id'], prio_sort_int, prio_sort_label, status, status_icon_name]
				# VIS_COL, ACT_COL, PRIO_COL, START_COL, DUE_COL, TAGS_COL, DESC_COL, PAGE_COL, TASKID_COL, PRIO_SORT_COL, PRIO_SORT_LABEL_COL, STATUS_COL, STATUS_ICON_NAME_COL
			modelrow[0] = self._filter_item(modelrow)
			myiter = self.real_model.append(parent_tree_iter, modelrow)

			if row['haschildren']:
				child_tasks = self.taskselection.list_tasks(row)
				self._append_tasks(child_tasks, myiter) # recurs

	def set_filter(self, string):
		# TODO allow more complex queries here - same parse as for search
		if string:
			inverse = False
			if string.lower().startswith('not '):
				# Quick HACK to support e.g. "not @waiting"
				inverse = True
				string = string[4:]
			self.filter = (inverse, string.strip().lower())
		else:
			self.filter = None
		self._eval_filter()

	def set_label_tag_filter(self, labels=None, tags=None, pages=None):
		if tags:
			self.tag_filter = [tag.lower() for tag in tags]
		else:
			self.tag_filter = None

		if labels:
			self.label_filter = [label.lower() for label in labels]
		else:
			self.label_filter = None

		if pages:
			self.page_filter = pages
		else:
			self.page_filter = None

		self._eval_filter()

	def _eval_filter(self):
		#logger.debug('Filtering with labels: %s tags: %s, filter: %s', self.label_filter, self.tag_filter, self.filter)
		if any((self.filter, self.tag_filter, self.label_filter, self.page_filter)):
			def filter(model, path, iter):
				visible = self._filter_item(model[iter])
				model[iter][VIS_COL] = visible
				if visible:
					parent = model.iter_parent(iter)
					while parent:
						model[parent][VIS_COL] = visible
						parent = model.iter_parent(parent)
		else:
			def filter(model, path, iter):
				model[iter][VIS_COL] = True

		self.real_model.foreach(filter)
		model = self.get_model()
		count = len(model) if model else 0
		self.emit('view-changed', count)
		self.expand_all()

	def _filter_item(self, modelrow):
		# This method filters case insensitive because both filters and
		# text are first converted to lower case text.
		visible = True

		pagename = modelrow[PAGE_COL].lower()
		description = modelrow[DESC_COL].lower()
		tags = [t.lower() for t in modelrow[TAGS_COL]]


		if visible and self.page_filter:
			pageparts = modelrow[PAGE_COL].split(':')
			visible = any(p in pageparts for p in self.page_filter)

		if visible and self.label_filter:
			# Any labels need to be present
			# (all does not make sense as they are mutual exclusive)
			for label in self.label_filter:
				if label in description:
					break
			else:
				visible = False # no label found

		if visible and self.tag_filter:
			# All tag should match
			if (_NO_TAGS in self.tag_filter and not tags) \
				or all(tag in tags for tag in self.tag_filter):
					visible = True
			else:
				visible = False

		if visible and self.filter:
			# And finally the filter string should match
			# FIXME: we are matching against markup text here - may fail for some cases
			inverse, string = self.filter
			if string.startswith('@'):
				match = string[1:].lower() in [t.lower() for t in tags]
			else:
				match = string in description or string in pagename
			if (not inverse and not match) or (inverse and match):
				visible = False

		return visible

	def do_focus_in_event(self, event):
		#print ">>>", self._today, datetime.date.today()
		if self._today != datetime.date.today():
			self.refresh()

	def do_row_activated(self, path, column):
		model = self.get_model()
		page = Path(model[path][PAGE_COL])
		text = self._get_raw_text(model[path])

		pageview = self.opener.open_page(page)
		pageview.show_find(text)

	def _get_raw_text(self, task):
		id = task[TASKID_COL]
		row = self.taskselection.get_task(id)
		return row['description']

	def do_initialize_popup(self, menu):
		item = Gtk.MenuItem.new_with_mnemonic(_('_Copy')) # T: menu label
		item.connect('activate', self.copy_to_clipboard)
		menu.append(item)
		self.populate_popup_expand_collapse(menu)


	def _query_tooltip_cb(self, widget, x, y, keyboard_tip, tooltip):
		context = widget.get_tooltip_context(x, y, keyboard_tip)
		if not context:
			return False

		model, iter = context.model, context.iter
		if not (model and iter):
			return

		task = model[iter][DESC_COL]
		start = model[iter][START_COL]
		due = model[iter][DUE_COL]
		page = model[iter][PAGE_COL]

		today = str(datetime.date.today())

		text = [task, '\n']
		if start and start > today:
			text += ['<b>', _('Start'), ':</b> ', start, '\n'] # T: start date for task
		if due != _MAX_DUE_DATE:
			text += ['<b>', _('Due'), ':</b> ', due, '\n'] # T: due date for task

		text += ['<b>', _('Page'), ':</b> ', encode_markup_text(page)] # T: page label

		tooltip.set_markup(''.join(text))
		return True

	def copy_to_clipboard(self, *a):
		'''Exports currently visible elements from the tasks list'''
		logger.debug('Exporting to clipboard current view of task list.')
		text = self.get_visible_data_as_csv()
		Clipboard.set_text(text)
			# TODO set as object that knows how to format as text / html / ..
			# unify with export hooks

	def get_visible_data_as_csv(self):
		text = ""
		for indent, prio, desc, due_date, start_date, page in self.get_visible_data():
			prio = str(prio)
			desc = decode_markup_text(desc)
			desc = '"' + desc.replace('"', '""') + '"'
			text += ",".join((prio, desc, due_date, start_date, page)) + "\n"
		return text

	def get_visible_data_as_html(self):
		html = '''\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html>
	<head>
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
		<title>Tasks</title>
		<meta name='Generator' content='Zim [%% zim.version %%]'>
		<style type='text/css'>
			table.tasklist {
				border-width: 1px;
				border-spacing: 2px;
				border-style: solid;
				border-color: gray;
				border-collapse: collapse;
			}
			table.tasklist th {
				border-width: 1px;
				padding: 1px;
				border-style: solid;
				border-color: gray;
			}
			table.tasklist td {
				border-width: 1px;
				padding: 1px;
				border-style: solid;
				border-color: gray;
			}
			.high {background-color: %s}
			.medium {background-color: %s}
			.alert {background-color: %s}
		</style>
	</head>
	<body>

<h1>Tasks</h1>

<table class="tasklist">
<tr><th>Prio</th><th>Task</th><th>Due</th><th>Start</th><th>Page</th></tr>
''' % (self.HIGH_COLOR, self.MEDIUM_COLOR, self.ALERT_COLOR)

		today = str(datetime.date.today())
		tomorrow = str(datetime.date.today() + datetime.timedelta(days=1))
		dayafter = str(datetime.date.today() + datetime.timedelta(days=2))
		for indent, prio, desc, due_date, start_date, page in self.get_visible_data():
			if prio >= 3:
					prio = '<td class="high">%s</td>' % prio
			elif prio == 2:
					prio = '<td class="medium">%s</td>' % prio
			elif prio == 1:
					prio = '<td class="alert">%s</td>' % prio
			else:
					prio = '<td>%s</td>' % prio

			if due_date and due_date <= today:
					due_date = '<td class="high">%s</td>' % due_date
			elif due_date == tomorrow:
					due_date = '<td class="medium">%s</td>' % due_date
			elif due_date == dayafter:
					due_date = '<td class="alert">%s</td>' % due_date
			else:
					due_date = '<td>%s</td>' % due_date

			start_date = '<td>%s</td>' % start_date

			desc = '<td>%s%s</td>' % ('&nbsp;' * (4 * indent), desc)
			page = '<td>%s</td>' % page

			html += '<tr>' + prio + desc + due_date + start_date + page + '</tr>\n'

		html += '''\
</table>

	</body>

</html>
'''
		return html

	def get_visible_data(self):
		rows = []

		def collect(model, path, iter):
			indent = len(path) - 1 # path is tuple with indexes

			row = model[iter]
			prio = row[PRIO_COL]
			desc = row[DESC_COL]
			due_date = row[DUE_COL]
			start_date = row[START_COL]
			page = row[PAGE_COL]

			if due_date == _MAX_DUE_DATE:
				due_date = ''

			if start_date == _MIN_START_DATE:
				start_date = ''

			rows.append((indent, prio, desc, due_date, start_date, page))

		model = self.get_model()
		model.foreach(collect)

		return rows




# Copyright 2010 Jaap Karssenberg <jaap.karssenberg@gmail.com>
#
# Inspired by and partially based on code from clac.py,
# Which is copyright 2009 Mark Borgerding and licensed under the GPL version 3


 # We are doing math in this module ...


import logging
import re
import math
import cmath

from zim.plugins import PluginClass
from zim.actions import action
from zim.errors import Error

from zim.gui.pageview import PageViewExtension


logger = logging.getLogger('zim.plugins.inlinecalculator')


# helper functions
def dip(x):
	'demote, if possible, a complex to scalar'
	if isinstance(x, complex) and x.imag == 0:
		return x.real
	else:
		return x

def which_call(x, mathfunc, cmathfunc, allowNegative=True):
	x = dip(x)
	if isinstance(x, complex) or (allowNegative == False and x < 0):
		return cmathfunc(x)
	else:
		return mathfunc(x)

# math functions defined here

def degrees(x):
	return x * 180 / math.pi

def radians(x):
	return x * math.pi / 180

def log(x, b=math.e):
	'log(x[, base]) -> the logarithm of x to the given base.\nIf the base not specified, returns the natural logarithm (base e) of x.'
	if isinstance(x, complex) or x < 0:
		return dip(cmath.log(x) / cmath.log(b))
	else:
		return math.log(x) / math.log(b)

def real(x):
	'return just the real portion'
	if isinstance(x, complex):
		return x.real
	else:
		return x

def imag(x):
	'return just the imaginary portion'
	if isinstance(x, complex):
		return x.imag
	else:
		return 0

def sign(x):
	'returns -1,0,1 for negative,zero,positive numbers'
	if x == 0:
		return 0
	elif x > 0:
		return 1
	else:
		return -1

def log2(x):
	'logarithm base 2'
	return log(x, 2)

def gcd(x, y):
	'greatest common denominator'
	while x > 0:
		(x, y) = (y % x, x) # Guido showed me this one on the geek cruise
	return y

def lcm(x, y):
	'least common multiple'
	return x * y / gcd(x, y)

def phase(z):
	'phase of a complex in radians'
	z = cpx(z)
	return math.atan2(z.imag, z.real)

def cpx(x):
	'convert a number or tuple to a complex'
	if isinstance(x, tuple):
		return complex(x[0], x[1])
	else:
		return complex(x)

def conj(x):
	'complex conjugate'
	x = cpx(x)
	return complex(x.real, -x.imag)

def complexify(x, func):
	'call func on the real and imaginary portions, creating a complex from the respective results'
	if isinstance(x, complex) and x.imag != 0:
		return dip(complex(func(x.real), func(x.imag)))
	else:
		return func(x)

# overwrite the built-in math functions that don't handle complex
def round(x):
	'nearest integer'
	if isinstance(x, complex):
		return complexify(x, round)
	else:
		return math.floor(x + .5)

def floor(x):
	'round towards negative infinity'
	return complexify(x, math.floor)

def ceil(x):
	'round towards positive infinity'
	return complexify(x, math.ceil)


# functions and constants available  within the safe eval construct
GLOBALS = {
	'__builtins__': None, # Don't allow open() etc.
	# builtins we want to keep
	'abs': abs,
	'ord': ord,
	'chr': chr,
	'hex': hex,
	'oct': oct,
	'int': int,
	# direct imports
	'e': math.e,
	'pi': math.pi,
	'atan2': math.atan2,
	'fmod': math.fmod,
	'frexp': math.frexp,
	'hypot': math.hypot,
	'ldexp': math.ldexp,
	'modf': math.modf,
	# other nice-to-have constants
	'j': cmath.sqrt(-1),
	# marshall between the math and cmath functions automatically
	'acos': lambda x: which_call(x, math.acos, cmath.acos),
	'asin': lambda x: which_call(x, math.asin, cmath.asin),
	'atan': lambda x: which_call(x, math.atan, cmath.atan),
	'cos': lambda x: which_call(x, math.cos, cmath.cos),
	'cosh': lambda x: which_call(x, math.cosh, cmath.cosh),
	'sin': lambda x: which_call(x, math.sin, cmath.sin),
	'sinh': lambda x: which_call(x, math.sinh, cmath.sinh),
	'tan': lambda x: which_call(x, math.tan, cmath.tan),
	'tanh': lambda x: which_call(x, math.tanh, cmath.tanh),
	'exp': lambda x: which_call(x, math.exp, cmath.exp),
	'log10': lambda x: which_call(x, math.log10, cmath.log10, False),
	'sqrt': lambda x: which_call(x, math.sqrt, cmath.sqrt, False),
	# functions defined here
	'degrees': degrees,
	'radians': radians,
	'log': log,
	'real': real,
	'imag': imag,
	'sign': sign,
	'log2': log2,
	'gcd': gcd,
	'lcm': lcm,
	'phase': phase,
	'conj': conj,
	'round': round,
	'floor': floor,
	'ceil': ceil,
	# synonyms
	'mag': abs,
	'angle': phase,
}


class ExpressionError(Error):

	description = _(
		'The inline calculator plugin was not able\n'
		'to evaluate the expression at the cursor.')
		# T: error description


_multiline_re = re.compile(r'--+\s+[+-]')
	# for multiline summation with "--- +" and similar


class InlineCalculatorPlugin(PluginClass):

	plugin_info = {
		'name': _('Inline Calculator'), # T: plugin name
		'description': _('''\
This plugin allows you to quickly evaluate simple
mathematical expressions in zim.

This is a core plugin shipping with zim.
'''), # T: plugin description
		'author': 'Jaap Karssenberg',
		'help': 'Plugins:Inline Calculator',
	}

	#~ plugin_preferences = (
		# key, type, label, default
	#~ )


	def process_text(self, text):
		'''Takes a piece of text and parses it for expressions
		to evaluate. Returns the text with result inserted or replaced.
		Will raise an exception on errors.
		'''
		# This method is separated from eval_math() for easy testing

		if '\n' in text:
			return self._process_multiline(text)
		else:
			return self._process_line(text)

	def _process_line(self, line):
		# Check for non-math prefix
		prefix = ''
		if ':' in line:
			i = line.rindex(':') + 1
			prefix = line[:i]
			line = line[i:]

		# Allow for chaining like "1 + 2 = 3 + 0.5 = 3.5" etc.
		if line.count('=') > 1:
			parts = line.split('=')
			prefix += '='.join(parts[:-2]) + '='
			line = '='.join(parts[-2:])

		# Check for whitespace postfix after previous answer
		postfix = ''
		stripped = line.rstrip()
		if '=' in line \
		and stripped != line and not stripped.endswith('='):
			i = len(line) - len(stripped)
			postfix = line[-i:]
			line = stripped

		# Strip previous answer and '='
		if '=' in line:
			i = line.index('=')
			line = line[:i]

		result = self.safe_eval(line)

		return prefix + line + '= ' + str(result) + postfix

	def _process_multiline(self, text):
		lines = text.splitlines()
		for i, line in enumerate(lines):
			if _multiline_re.match(line):
				operator = line.strip()[-1]
				break
		else:
			raise ExpressionError(_('Could not parse expression'))
				# T: error message

		sep = ' %s ' % operator
		expression = sep.join('(%s)' % l for l in lines[:i])
		result = self.safe_eval(expression)

		lines = lines[:i + 1] + [str(result)]
		return '\n'.join(lines) + '\n'


	def safe_eval(self, expression):
		'''Safe evaluation of a python expression'''
		try:
			return eval(expression, GLOBALS, {})
		except Exception as error:
			msg = '%s: %s' % (error.__class__.__name__, error)
			raise ExpressionError(msg)


class InlineCalculatorPageViewExtension(PageViewExtension):

	@action(_('Evaluate _Math')) # T: menu item
	def eval_math(self):
		'''Action called by the menu item or key binding,
		will look at the cursor for an expression to evaluate.
		'''
		buffer = self.pageview.textview.get_buffer()
			# XXX- way to long chain of objects here

		# FIXME: what do we do for selections ?

		cursor = buffer.get_iter_at_mark(buffer.get_insert())
		start, end = buffer.get_line_bounds(cursor.get_line())
		line = start.get_text(end)

		if not line or line.isspace():
			# Empty line, look at previous line
			if cursor.get_line() > 0:
				start, end = buffer.get_line_bounds(cursor.get_line() - 1)
				cursor = end.copy()
				cursor.backward_char()
				line = start.get_text(end)
			else:
				return # silent fail

		if _multiline_re.match(line):
			# Search for start of block - iterate back to empty line
			lineno = cursor.get_line()
			while lineno >= 0:
				mystart, myend = buffer.get_line_bounds(lineno)
				myline = mystart.get_text(myend)
				if not myline or myline.isspace():
					break
				else:
					start = mystart
					lineno -= 1
		else:
			# One line expression, just pass it on
			# FIXME skip forward past next word if any if last char is '='
			end = cursor

		orig = start.get_text(end)
		logger.debug('Inline calculator eval: >%s<', orig)
		new = self.plugin.process_text(orig)
		with buffer.user_action:
			buffer.delete(start, end)
			buffer.insert(start, new)




# Copyright 2013-2016 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''
This module is responsible for commandline parsing. It provides a base
class for the commandline commands defined in L{zim.main}
'''

import os

from getopt import gnu_getopt, GetoptError

import logging

logger = logging.getLogger('zim')


from zim.errors import Error


class UsageError(Error):
	'''Error raised when commands do not have correct
	number or type of arguments
	'''
	pass


class Command(object):
	'''Base class for commandline commands, used by zim to abstract
	part of the C{main()} functionality and allow better testability
	of commandline arguments.

	Sub-classes can define the options and arguments that they require.
	Then only the C{run()} method needs to be defined to implement the
	actual command. In the C{run()} method C{self.opts} and C{self.args}
	can be accessed to get the commandline options (dict) and the
	commandline arguments (list) respectively.
	'''

	arguments = () #: Define arguments, e.g ('NOTEBOOK', '[PAGE]')

	options = () #: Define options by 3-tuple of long, short & description.
		#: E.g. ("foo=", "f", "set parameter for foo")
		#: For options that can appear multiple times,
		#: assign a list "[]" in "self.opts" before parse_options is called

	default_options = (
		('verbose', 'V', 'Verbose output'),
		('debug', 'D', 'Debug output'),
		('help', 'h', 'Print help text and exit'),
	) #: Default options for all commands

	cmdhelp = '' #: If defined, this text is printed on --help for this command

	def __init__(self, command, pwd=None):
		'''Constructor
		@param command: the command switch (first commandline argument)
		@param pwd: optional working directory path
		'''
		self.command = command
		self.args = []
		self.opts = {}
		self.pwd = pwd or os.getcwd()

	def parse_options(self, *args):
		'''Parse commandline options for this command
		Sets the attributes 'args' and 'opts' to a list of arguments
		and a dictionary of options respectively
		@param args: all remaining options to be parsed
		@raises GetOptError: when options are not correct
		'''
		options = ''
		long_options = []
		options_map = {}
		requires_arg = set()
		for l, s, desc in self.default_options + self.options:
			long_options.append(l)
			if l.endswith('='):
				requires_arg.add(l.strip('='))

			if s and l.endswith('='):
				options += s + ':'
				options_map[s] = l.strip('=')
			elif s:
				options += s
				options_map[s] = l

		optlist, args = gnu_getopt(args, options, long_options)
		self.args += args

		for o, a in optlist:
			key = o.strip('-')
			key = options_map.get(key, key)
			if a == '' and key not in requires_arg: # implies boolean flag
				self.opts[key] = True
			elif key in self.opts and isinstance(self.opts[key], list):
				self.opts[key].append(a)
			else:
				self.opts[key] = a

	def get_options(self, *names):
		'''Retrieve a dict with a sub-set of the command options
		@param names: that options in the subset
		'''
		return dict((k, self.opts.get(k)) for k in names)

	def get_arguments(self):
		'''Get the arguments, to be used by the implementation of C{run()}
		@raises UsageError: when arguments are not correct
		@returns: tuple of arguments, padded with None to correct length
		'''
		minimum = len([a for a in self.arguments if not a.startswith('[')])
		if len(self.args) < minimum:
			raise UsageError('Command %s takes %i arguments' % (self.command, minimum))
		elif len(self.args) > len(self.arguments) and not self.arguments[-1].endswith('+'):
			raise UsageError('Command %s takes only %i arguments' % (self.command, len(self.args)))
		else:
			return tuple(self.args) \
				+ (None,) * (len(self.arguments) - len(self.args))

	def ignore_options(self, *options):
		for option in options:
			if self.opts.get(option) is not None:
				logger.warning('Option "%s" is ignored for this command', option)

	def run(self):
		'''Run the command
		@raises UsageError: when arguments are not correct
		@implementation: must be implemented by subclasses
		'''
		raise NotImplementedError


class GtkCommand(Command):
	'''Base class for commandline commands that result in a Gtk
	user interface being presented to the user or want to interact with the
	Gtk user interface.

	If the C{run()} method returns a window, it will be added to the
	application top level windows. And a C{Gtk.main} loop will run until
	all windows are destroyed.

	Commands derived from this class can be dispatched to the main application
	process.

	NOTE: Do _not_ call C{Gtk.main} from the command, this will be
	done by the application object.
	'''

	default_options = Command.default_options + (
		('non-unique', '', 'start a new process, do not connect to an existing process'),
		('standalone', '', 'start a new process per notebook, implies --non-unique'),
	)

	def handle_local_commandline(self, args):
		'''Method called in local process before we (try to) dispatch the
		command to the primary process
		If the command is called from the primary process directly, this method
		gets never called.
		Since the only communication from the local to the primary process is
		by passing on the commandline arguments, the main function of this method
		is to modify the commandline arguments.
		@param args: all remaining options
		@returns: modified command line
		'''
		return args



# Copyright 2013-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

# This class is in a seperate module to avoid Gtk imports in the main module
# imports are a bit circular

import os
import logging
import signal
from typing import List, Optional

logger = logging.getLogger('zim')

import zim.errors

from gi.repository import Gio
from gi.repository import Gtk

from . import _application_startup, build_command, UsageError


class ZimGtkApplication(Gtk.Application):
	
	APPLICATION_ID = 'org.zim_wiki.Zim' # For practical reasons use "_" instead of "-"
	
	def __init__(self, non_unique=False, standalone=False):
		'''Constructor
		@param non_unique: do not attempt to run a unique process but start a new process here
		@param standalone: in standalone mode, each notebook is opened in it's own process, 
		this implies C{non_unique}
		'''
		Gtk.Application.__init__(self)
		self.set_application_id(self.APPLICATION_ID)
		self.standalone = standalone
		if standalone or non_unique:
			self.set_flags(Gio.ApplicationFlags.HANDLES_COMMAND_LINE|Gio.ApplicationFlags.NON_UNIQUE)
		else:
			self.set_flags(Gio.ApplicationFlags.HANDLES_COMMAND_LINE)
		self.connect('startup', self.__class__._do_startup) # Hack to avoid bug in python-gtk interaction
		self.connect('shutdown', self.__class__._do_shutdown)

	def _do_startup(self):
		_application_startup()

		def handle_sigterm(signal, frame):
			logger.info('Got SIGTERM, quit')
			self.quit()

		signal.signal(signal.SIGTERM, handle_sigterm)

		from zim.gui.widgets import gtk_window_set_default_icon
		gtk_window_set_default_icon()

		zim.errors.set_use_gtk(True)

	def _do_shutdown(self):
		# MainWindow implements saving in the destroy method
		for window in self.get_windows():
			window.destroy()

	def do_command_line(self, gcommandline):
		# Handler in primary process to process commandline and start application
		# if started from remote process, the exit code is given back to that process
		# and it exits while primary process keeps running as long as there are windows
		args = gcommandline.get_arguments()
		if 'python' in os.path.basename(args[0]):
			args = args[2:] # strip python interpreter and zim script name
		else:
			args = args[1:] # strip zim executable name
		return self.run_commandline(args, gcommandline.get_cwd())

	def run_commandline(self, args: List[str], pwd: Optional[str] = None) -> int:
		'''Run a commandline in the current process
		@param args: the commandline options
		@param pwd: the working directory as string path
		@returns: exit value as integer, 0 is OK, other values for errors
		'''
		try:
			args = [a for a in args if a is not None]
			cmd = build_command(args, pwd=pwd)
			window = cmd.run()
		except Exception as err:
			import sys
			print(err, file=sys.stderr)
			return 1
		else:
			if window and isinstance(window, Gtk.Window):
				self.add_window(window)
				window.present()
			elif window:
				# E.g. trayicon, not a window, but we need to hold application
				self.hold()
				window.connect('destroy', lambda *a: self.release())

			return 0

	def open_notebook(self, location, pagelink=None):
		'''Open a notebook either in an existing or a new window
		@param location: notebook location as uri or object with "uri" attribute like C{File}, C{Notebook}, or C{NotebookInfo}
		@param pagelink: optional page link (including optional anchor ID) as string or C{Path} object
		'''
		uri = location if isinstance(location, str) else location.uri
		pagelink = pagelink if isinstance(pagelink, (str, type(None))) else pagelink.name

		if self.standalone:
			self._spawn_new_instance('--gui', uri, pagelink)
		else:
			self.run_commandline(['--gui', uri, pagelink])

	def open_manual(self, pagelink=None):
		'''Open the manual either in an existing or a new window
		@param pagelink: optional page link (including optional anchor ID) as string or C{Path} object
		'''
		pagelink = pagelink if isinstance(pagelink, (str, type(None))) else pagelink.name

		if self.standalone:
			self._spawn_new_instance('--manual', pagelink)
		else:
			self.run_commandline(['--manual', pagelink])

	def _spawn_new_instance(self, *arg):
		# This method forces a new zim process to be run
		# used for "standalone" mode where each notebook gets
		# it's own process
		from zim import ZIM_EXECUTABLE
		from zim.applications import Application

		args = [ZIM_EXECUTABLE] + [a for a in arg if a is not None]
		if not '--standalone' in args:
			args.append('--standalone')

		# more detailed logging has lower number, so WARN > INFO > DEBUG
		loglevel = logging.getLogger().getEffectiveLevel()
		if loglevel <= logging.DEBUG:
			args.append('-D',)
		elif loglevel <= logging.INFO:
			args.append('-V',)

		Application(args).spawn()




# Copyright 2012-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''Module with assorted useful classes and functions used in the zim code'''


import collections


class classproperty(object):
	'''Like C{property()} but for klass properties
	Typically used as decorator
	'''

	def __init__(self, func):
		self.func = func

	def __get__(self, obj, owner):
		return self.func(owner)


class LastDefinedOrderedDict(collections.OrderedDict):
	'''Class that behaves like a dict but keeps items the order they were defined.
	Updating an items keeps it at the current position, removing and
	re-inserting an item puts it at the end of the sequence.
	'''

	def __setitem__(self, key, value):
		if not key in self:
			super().__setitem__(key, value)
			self.move_to_end(key)
		else:
			super().__setitem__(key, value)


class MovingWindowIter(object):
	'''Iterator yields a 3-tuple of the previous item, the current item
	and the next item while iterating a give iterator.
	Previous or next item will be C{None} if not available.
	Use as:

		for prev, current, next in MovingWindowIter(mylist):
			....

	@ivar items: current 3-tuple
	@ivar last: C{True} if we are at the last item
	'''

	def __init__(self, iterable):
		self._iter = iter(iterable)
		try:
			first = next(self._iter)
		except StopIteration:
			# empty list
			self.last = True
			self.last = (None, None, None)
		else:
			self.last = False
			self.items = (None, None, first)

	def __iter__(self):
		return self

	def __next__(self):
		if self.last:
			raise StopIteration

		discard, prev, current = self.items
		try:
			mynext = next(self._iter)
		except StopIteration:
			self.last = True
			self.items = (prev, current, None)
		else:
			self.items = (prev, current, mynext)

		return self.items



#
# ploteditor.py
#
# This is a plugin for Zim, which allows inserting GNU R scripts to
# have Zim generate plots from them.
#
# Author: Lee Braiden <lee.b@irukado.org>
# Date: 2010-03-13
# Copyright (c) 2010, released under the GNU GPL v2 or higher
#
# Heavily based on equationeditor.py plugin as of:
# bzr revno 212, (2010-03-10), marked as
# Copyright 2009 Jaap Karssenberg <jaap.karssenberg@gmail.com>
#

import glob
import re

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import \
	ImageGeneratorClass, BackwardImageGeneratorObjectType

from zim.newfs import LocalFile, TmpFile
from zim.config import data_file
from zim.templates import get_template
from zim.applications import Application

# TODO put these commands in preferences
gnu_r_cmd = ('R',)


class InsertGNURPlotPlugin(PluginClass):

	plugin_info = {
		'name': _('Insert GNU R Plot'), # T: plugin name
		'description': _('''\
This plugin provides a plot editor for zim based on GNU R.
'''), # T: plugin description
		'help': 'Plugins:GNU R Plot Editor',
		'author': 'Lee Braiden',
	}

	@classmethod
	def check_dependencies(klass):
		has_gnur = Application(gnu_r_cmd).tryexec()
		return has_gnur, [('GNU R', has_gnur, True)]


class BackwardGnuRPlotImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+gnu_r_plot'
	label = _('GNU R Plot') # T: menu item
	syntax = 'r'
	scriptname = 'gnu_r_plot.r'


class GNURPlotGenerator(ImageGeneratorClass):

	imagefile_extension = '.png'

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.template = get_template('plugins', 'gnu_r_editor.r')
		self.plotscriptfile = TmpFile('gnu_r_plot.r')

	def generate_image(self, text):
		plotscriptfile = self.plotscriptfile
		pngfile = LocalFile(plotscriptfile.path[:-2] + '.png')

		plot_width = 480 # default image width (px)
		plot_height = 480 # default image height (px)

		# LOOK for image size in comments of the script
		r = re.search(r"^#\s*WIDTH\s*=\s*([0-9]+)$", text, re.M)
		if r:
			plot_width = int(r.group(1))
		r = re.search(r"^#\s*HEIGHT\s*=\s*([0-9]+)$", text, re.M)
		if r:
			plot_height = int(r.group(1))

		template_vars = {
			'gnu_r_plot_script': text,
			'r_width': plot_width,
			'r_height': plot_height,
			'png_fname': pngfile.path.replace('\\', '/'),
				# Even on windows, GNU R expects unix path seperator
		}

		# Write to tmp file usign the template for the header / footer
		lines = []
		self.template.process(lines, template_vars)
		plotscriptfile.writelines(lines)
		#print '>>>%s<<<' % plotscriptfile.read()

		# Call GNU R
		try:
			gnu_r = Application(gnu_r_cmd)
			gnu_r.run(args=('-f', plotscriptfile.basename, '--vanilla'), cwd=plotscriptfile.parent())
		except:
			return None, None # Sorry, no log
		else:
			return pngfile, None

	def cleanup(self):
		path = self.plotscriptfile.path
		for path in glob.glob(path[:-2] + '.*'):
			LocalFile(path).remove()



# Copyright 2012-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''Functions for localized "natural" sorting'''


import locale
import re
import unicodedata


_num_re = re.compile(r'\d+')


def natural_sort_key(string, numeric_padding=5):
	'''Format string such that it gives 'natural' sorting on string
	compare. Will pad any numbers in the string with "0" such that "10"
	sorts after "9". Also includes C{locale.strxfrm()}.

	@note: sorting not 100% stable for case, so order between "foo" and
	"Foo" is not defined. For this reason when sort needs to be absolutely
	stable it is advised to sort based on tuples of
	C{(sort_key, original_string)}. Or use either L{natural_sort()} or
	L{natural_sorted()} instead.

	@param string: the string to format
	@param numeric_padding: number of digits to use for padding
	@returns: string transformed to sorting key
	'''
	templ = '%0' + str(numeric_padding) + 'i'
	string = string or ''  # Handle None by sorting as empty string
	string = unicodedata.normalize('NFC', string.strip())
	string = _num_re.sub(lambda m: templ % int(m.group()), string)
	string = string.lower() # sort case insensitive

	try:
		bytestring = locale.strxfrm(string)
			# 8-bit byte string - enode to hex -- in pyton3 check if byte data type is handled better by sqlite3 and others
	except MemoryError:
		# Known python issue :(
		bytestring = string

	key = ''.join(["%02x" % ord(c) for c in bytestring])
	return key


def natural_sort(list, key=None):
	'''Natural sort a list in place.
	See L{natural_sort_key} for details.
	@param list: list of strings to be sorted
	@param key: function producing strings for list items
	'''
	if key:
		def func(s):
			s = key(s)
			return (natural_sort_key(s), s)
	else:
		func = lambda s: (natural_sort_key(s), s)
	list.sort(key=func)


def natural_sorted(iter, key=None):
	'''Natural sort a list.
	See L{natural_sort_key} for details.
	@param iter: list or iterable of strings to be sorted
	@param key: function producing strings for list items
	@returns: sorted copy of the list
	'''
	l = list(iter) # cast to list and implicit copy
	natural_sort(l, key=key)
	return l



# Copyright 2008-2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>


from gi.repository import Gdk


from zim.formats import \
	BULLET, CHECKED_BOX, UNCHECKED_BOX, XCHECKED_BOX, TRANSMIGRATED_BOX, \
	MIGRATED_BOX, LINE, OBJECT, HEADING, LISTITEM, BLOCK_LEVEL, FORMATTEDTEXT, \
	LINK, TAG, ANCHOR, EMPHASIS, STRONG, MARK, STRIKE, SUBSCRIPT, SUPERSCRIPT, IMAGE, \
	VERBATIM, VERBATIM_BLOCK, \
	BULLETLIST, NUMBEREDLIST, BLOCK

# Special character that acts as placeholder for images and objects
PIXBUF_CHR = '\uFFFC'

# Special character that acts as placeholder for images and objects
PIXBUF_CHR = '\uFFFC'


# Tag names used internally but not part of tokenlist for formats
HEADING_1, HEADING_2, HEADING_3, HEADING_4, HEADING_5, HEADING_6 = 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'
HEADING_1_to_6 = (HEADING_1, HEADING_2, HEADING_3, HEADING_4, HEADING_5, HEADING_6)
PAGE_LINK = 'page-link'


# Pixbuf types used internally
ICON = 'icon'


## Bullets and checkboxes

STOCK_CHECKED_BOX = 'zim-checked-box'
STOCK_UNCHECKED_BOX = 'zim-unchecked-box'
STOCK_XCHECKED_BOX = 'zim-xchecked-box'
STOCK_MIGRATED_BOX = 'zim-migrated-box'
STOCK_TRANSMIGRATED_BOX = 'zim-transmigrated-box'

BULLET_TYPES = {
	CHECKED_BOX: STOCK_CHECKED_BOX,
	UNCHECKED_BOX: STOCK_UNCHECKED_BOX,
	XCHECKED_BOX: STOCK_XCHECKED_BOX,
	MIGRATED_BOX: STOCK_MIGRATED_BOX,
	TRANSMIGRATED_BOX: STOCK_TRANSMIGRATED_BOX,
}

BULLETS_FROM_STOCK = {}
for _b in BULLET_TYPES:
	BULLETS_FROM_STOCK[BULLET_TYPES[_b]] = _b

AUTOFORMAT_BULLETS = {
	'*': BULLET,
	'[]': UNCHECKED_BOX,
	'[ ]': UNCHECKED_BOX,
	'[*]': CHECKED_BOX,
	'[x]': XCHECKED_BOX,
	'[>]': MIGRATED_BOX,
	'[<]': TRANSMIGRATED_BOX,
	'()': UNCHECKED_BOX,
	'( )': UNCHECKED_BOX,
	'(*)': CHECKED_BOX,
	'(x)': XCHECKED_BOX,
	'(>)': MIGRATED_BOX,
	'(<)': TRANSMIGRATED_BOX,
}

BULLETS = (BULLET, UNCHECKED_BOX, CHECKED_BOX, XCHECKED_BOX, MIGRATED_BOX, TRANSMIGRATED_BOX)
CHECKBOXES = (UNCHECKED_BOX, CHECKED_BOX, XCHECKED_BOX, MIGRATED_BOX, TRANSMIGRATED_BOX)

NUMBER_BULLET = '#.' # Special case for autonumbering

# Styles
INDENT_BLOCK_STYLE = 'indent'
BULLET_LIST_STYLE = 'bullet-list'
NUMBERED_LIST_STYLE = 'numbered-list'


## Keybindings

# Check the (undocumented) list of constants in Gtk.keysyms to see all names
KEYVALS_HOME = list(map(Gdk.keyval_from_name, ('Home', 'KP_Home')))
KEYVALS_ENTER = list(map(Gdk.keyval_from_name, ('Return', 'KP_Enter', 'ISO_Enter')))
KEYVALS_BACKSPACE = list(map(Gdk.keyval_from_name, ('BackSpace',)))
KEYVALS_TAB = list(map(Gdk.keyval_from_name, ('Tab', 'KP_Tab')))
KEYVALS_LEFT_TAB = list(map(Gdk.keyval_from_name, ('ISO_Left_Tab',)))

# ~ CHARS_END_OF_WORD = (' ', ')', '>', '.', '!', '?')
CHARS_END_OF_WORD = ('\t', ' ', ')', '>', ';')
KEYVALS_END_OF_WORD = list(map(
	Gdk.unicode_to_keyval, list(map(ord, CHARS_END_OF_WORD)))) + KEYVALS_TAB

KEYVALS_ASTERISK = (
	Gdk.unicode_to_keyval(ord('*')), Gdk.keyval_from_name('KP_Multiply'))
KEYVALS_SLASH = (
	Gdk.unicode_to_keyval(ord('/')), Gdk.keyval_from_name('KP_Divide'))
KEYVALS_GT = (Gdk.unicode_to_keyval(ord('>')),)
KEYVALS_SPACE = (Gdk.unicode_to_keyval(ord(' ')),)

KEYVAL_ESC = Gdk.keyval_from_name('Escape')
KEYVAL_POUND = Gdk.unicode_to_keyval(ord('#'))

# States that influence keybindings - we use this to explicitly
# exclude other states. E.g. MOD2_MASK seems to be set when either
# numlock or fn keys are active, resulting in keybindings failing
KEYSTATES = Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.META_MASK | Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.MOD1_MASK

# Minimal distance from mark to window border after scroll_to_mark()
SCROLL_TO_MARK_MARGIN = 0.2




# Copyright 2008-2015 Jaap Karssenberg <jaap.karssenberg@gmail.com>


import re
import logging
import itertools

from typing import Generator, Generic, List, Optional, Union

logger = logging.getLogger('zim.notebook')


from zim.parse.links import link_type
from zim.errors import Error

import zim.formats
import zim.newfs

from zim.signals import SignalEmitter, SIGNAL_NORMAL

import zim.datetimetz as datetime


_pagename_reduce_colon_re = re.compile('::+')
_pagename_invalid_char_re = re.compile(
	'(' +
		r'^[_\W]+|(?<=:)[_\W]+' +
	'|' +
		'[' + re.escape(''.join(
			("?", "#", "/", "\\", "*", '"', "<", ">", "|", "%", "\t", "\n", "\r")
		)) + ']' +
	')',
re.UNICODE)
	# This pattern matches a non-alphanumber at start or after the ':'
	# separator. It also matches any invalid character.
	# The UNICODE flag is used to make the alphanumber check international.


def shortest_unique_names(paths: List['Path']) -> List[str]:
	'''Returns the shortest unique name for each path in paths
	@param paths: list of L{Path} objects
	@returns: list of strings
	'''
	by_basename = {}
	for path in paths:
		basename = path.basename
		mylist = by_basename.setdefault(basename, [])
		mylist.append(path)

	result = []
	for path in paths:
		basename = path.basename
		conflicts = by_basename[basename]
		if len(conflicts) == 1:
			result.append(path.basename)
		else:
			conflicts.remove(path)
			conflicts.insert(0, path) # shuffle path of interest to front
			reverse_paths = [reversed(p.name.split(':')) for p in conflicts]
			names = []
			for parts in itertools.zip_longest(*reverse_paths):
				if parts[0] is None:
					break
				elif parts[0] not in parts[1:]:
					names.append(parts[0])
					break
				else:
					names.append(parts[0])

			result.append(':'.join(reversed(names)))

	return result


class Path():
	'''Class representing a page name in the notebook

	This is the parent class for the Page class. It contains the name
	of the page and is used instead of the actual page object by methods
	that only need to know the name of the page. Path objects have no
	internal state and are essentially normalized page names. It also
	has a number of methods to compare page names and determining what
	the parent pages are etc.

	@ivar name: the full name of the path
	@ivar parts: all the parts of the name (split on ":")
	@ivar basename: the basename of the path (last part of the name)
	@ivar namespace: the name for the parent page or empty string
	@ivar isroot: C{True} when this Path represents the top level namespace
	@ivar parent: the L{Path} object for the parent page


	Valid characters in page names
	==============================

	A number of characters are not valid in page names as used in Zim
	notebooks.

	Reserved characters are:
	  - The ':' is reserved as separator
	  - The '?' is reserved to encode url style options
	  - The '#' is reserved as anchor separator
	  - The '/' and '\' are reserved to distinguish file links & urls
	  - First character of each part MUST be alphanumeric
		(including utf8 letters / numbers)

	For file system filenames we can not use:
	'\', '/', ':', '*', '?', '"', '<', '>', '|'
	(checked both win32 & posix)

	Do not allow '\n' and '\t' for obvious reasons

	Allowing '%' will cause problems with sql wildcards sooner
	or later - also for url decoding ambiguity it is better to
	keep this one reserved.

	All other characters are allowed in page names

	Note that Zim version < 0.42 used different rules that are not
	fully compatible, this is important when upgrading old notebooks.
	See L{Notebook.cleanup_pathname_zim028()}
	'''

	__slots__ = ('name',)

	@staticmethod
	def assertValidPageName(name: str) -> None:
		'''Raises an C{AssertionError} if C{name} does not represent
		a valid page name.
		This is a strict check, most names that fail this test can still
		be cleaned up by the L{makeValidPageName()}.
		@raises AssertionError: if the name is not valid
		'''
		assert isinstance(name, str)
		if not name.strip(':') \
		or _pagename_reduce_colon_re.search(name) \
		or _pagename_invalid_char_re.search(name):
			raise AssertionError('Not a valid page name: %s' % name)

	@staticmethod
	def makeValidPageName(name: str) -> str:
		'''Remove any invalid characters from the string and return
		a valid page name. Only string that can not be turned in
		somthing valid is a string that reduces to an empty string
		after removing all invalid characters.
		@raises ValueError: when the result would be an empty string
		'''
		newname = _pagename_reduce_colon_re.sub(':', name.strip(':'))
		newname = _pagename_invalid_char_re.sub('', newname)
		newname = newname.replace('_', ' ')
		try:
			Path.assertValidPageName(newname)
		except AssertionError:
			raise ValueError('Not a valid page name: %s (was: %s)' % (newname, name))
		return newname

	def __init__(self, name: Union[str, tuple]):
		'''Constructor.

		@param name: the absolute page name in the right case as a
		string or as a tuple strings

		The name ":" is used as a special case to construct a path for
		the toplevel namespace in a notebook.

		@note: This constructor does not do any checks for the sanity of
		the path name. Never construct a path directly from user input,
		but use either L{index.lookup_from_user_input()} or first check the
		name with L{makeValidPageName()}
		'''
		if isinstance(name, (list, tuple)):
			self.name = ':'.join(name)
		else:
			self.name = name.strip(':')

		try:
			self.name = str(self.name)
		except UnicodeDecodeError:
			raise ValueError('BUG: invalid input, page names should be in ascii, or given as unicode')

	@classmethod
	def new_from_zim_config(klass, string: str) -> 'Path':
		'''Returns a new object based on the string representation for
		that path.
		'''
		return klass(klass.makeValidPageName(string))

	def serialize_zim_config(self) -> str:
		'''Returns the name for serializing this path'''
		return self.name

	def __repr__(self):
		return '<%s: %s>' % (self.__class__.__name__, self.name)

	def __str__(self):
		return self.name

	def __hash__(self):
		return self.name.__hash__()

	def __eq__(self, other):
		'''Paths are equal when their names are the same'''
		if isinstance(other, Path):
			return self.name == other.name
		else: # e.g. path == None
			return False

	def __ne__(self, other):
		'''Paths are not equal when their names are not the same'''
		return not self.__eq__(other)

	def __add__(self, name):
		'''C{path + name} is an alias for C{path.child(name)}'''
		return self.child(name)

	@property
	def parts(self) -> List[str]:
		'''Get all the parts of the name (split on ":")'''
		return self.name.split(':')

	@property
	def basename(self) -> str:
		'''Get the basename of the path (last part of the name)'''
		i = self.name.rfind(':') + 1
		return self.name[i:]

	@property
	def namespace(self) -> str:
		'''Gives the name for the parent page.
		Returns an empty string for the top level namespace.
		'''
		i = self.name.rfind(':')
		if i > 0:
			return self.name[:i]
		else:
			return ''

	@property
	def isroot(self) -> bool:
		'''C{True} when this Path represents the top level namespace'''
		return self.name == ''

	def relname(self, path: 'Path') -> str: # TODO make this use HRef !
		'''Get a part of this path relative to a parent path

		@param path: a parent L{Path}

		@raises ValueError: if C{path} is not a parent

		@returns: the part of the path that is relative to C{path}
		'''
		if path.name == '': # root path
			return self.name
		elif self.name.startswith(path.name + ':'):
			i = len(path.name) + 1
			return self.name[i:].strip(':')
		else:
			raise ValueError('"%s" is not below "%s"' % (self, path))

	@property
	def parent(self) -> 'Path':
		'''Get the path for the parent page'''
		namespace = self.namespace
		if namespace:
			return Path(namespace)
		elif self.isroot:
			return None
		else:
			return Path(':')

	def parents(self) -> Generator['Path', None, None]:
		'''Generator function for parent Paths including root'''
		if ':' in self.name:
			path = self.name.split(':')
			path.pop()
			while len(path) > 0:
				namespace = ':'.join(path)
				yield Path(namespace)
				path.pop()
		yield Path(':')

	def child(self, basename: str) -> 'Path':
		'''Get a child Path

		@param basename: the relative name for the child
		@returns: a new L{Path} object
		'''
		return Path(self.name + ':' + basename)

	def ischild(self, parent: 'Path') -> bool:
		'''Check whether this path is a child of a given path
		@param parent: a L{Path} object
		@returns: True when this path is a (grand-)child of C{parent}
		'''
		return parent.isroot or self.name.startswith(parent.name + ':')

	def match_namespace(self, namespace: 'Path') -> bool:
		'''Check whether this path is in a specific section of the notebook
		@param namespace: a L{Path} object
		@returns: True when this path is equal to C{namespace} or is a (grand-)child of C{namespace}
		'''
		return namespace.isroot or self.name == namespace.name or self.name.startswith(namespace.name + ':')

	def commonparent(self, other: 'Path') -> 'Path':
		'''Find a common parent for two Paths

		@param other: another L{Path} object

		@returns: a L{Path} object for the first common parent
		'''
		parent = []
		parts = self.parts
		other = other.parts
		if parts[0] != other[0]:
			return Path(':') # root
		else:
			for i in range(min(len(parts), len(other))):
				if parts[i] == other[i]:
					parent.append(parts[i])
				else:
					return Path(':'.join(parent))
			else:
				return Path(':'.join(parent))


HRef_rel_flavor = int
HREF_REL_ABSOLUTE = 0
HREF_REL_FLOATING = 1
HREF_REL_RELATIVE = 2

class HRef():
	'''Represents a link as it appears in the wiki source.
	Contains semantic information about the link type, but
	does not contain an end point.

	Introduced to help preserve link type when moving content.

	@note: To create the shortest link between two pages,
	use L{Notebook.pages.create_link}

	@note: To resolve a link into a L{Path}, create a L{HRef}
	and pass it to L{Notebook.pages.resolve_link}.
	'''

	__slots__ = ('rel', 'names', 'anchor')

	@classmethod
	def makeValidHRefString(klass, href: str) -> str:
		return klass.new_from_wiki_link(href).to_wiki_link()

	@classmethod
	def new_from_wiki_link(klass, href: str) -> 'HRef':
		'''Constructor that constructs a L{HRef} object for a link as
		written in zim's wiki syntax.
		@param href: a string for the link
		@raises ValueError: when the string could not be parsed
		(see L{Path.makeValidPageName()})

		@note: This method HRef class assumes the logic of our wiki links.
		For other formats, a separate constructor may be needed.
		'''
		href = href.strip()

		if href.startswith(':'):
			rel = HREF_REL_ABSOLUTE
		elif href.startswith('+'):
			rel = HREF_REL_RELATIVE
		else:
			rel = HREF_REL_FLOATING

		anchor = None
		if '#' in href:
			href, anchor = href.split('#', 1)
			anchor = zim.formats.heading_to_anchor(anchor) # make valid achor string

		names = Path.makeValidPageName(href.lstrip('+')) if href else ""

		return klass(rel, names, anchor)

	def __init__(self, rel: HRef_rel_flavor, names: str, anchor: Optional[str] = None):
		self.rel = rel
		self.names = names
		self.anchor = anchor

	def __str__(self):
		rel = {HREF_REL_ABSOLUTE: 'abs', HREF_REL_FLOATING: 'float', HREF_REL_RELATIVE: 'rel'}[self.rel]
		return '<%s: %s %s %s>' % (self.__class__.__name__, rel, self.names, self.anchor)

	def __eq__(self, other):
		return (self.__class__ is other.__class__) and (str(self) == str(other))

	def parts(self) -> Union[List, List[str]]:
		return self.names.split(':') if self.names else []

	def short_name(self) -> str:
		" Returns the last name part and/or anchor if any. "
		name = self.parts()[-1] if self.names else ""
		return name + "#" + self.anchor if self.anchor else name

	def to_wiki_link(self) -> str:
		'''Returns href as text for wiki link'''
		if self.rel == HREF_REL_ABSOLUTE:
			link = ":" + self.names.strip(':')
		elif self.rel == HREF_REL_RELATIVE:
			link = "+" + self.names
		else:
			link = self.names

		if self.anchor:
			link += "#" + self.anchor

		return link


class PageError(Error):

	def __init__(self, path):
		self.path = path
		self.msg = self._msg % path.name


class PageReadOnlyError(PageError):
	_msg = _('Can not modify page: %s') # T: error message for read-only pages


class Page(Path, SignalEmitter):
	'''Class to represent a single page in the notebook.

	Page objects inherit from L{Path} but have internal state reflecting
	content in the notebook. We try to keep Page objects unique
	by hashing them in L{Notebook.get_page()}, Path object on the other
	hand are cheap and can have multiple instances for the same logical path.
	We ask for a path object instead of a name in the constructor to
	encourage the use of Path objects over passing around page names as
	string.

	You can use a Page object instead of a Path anywhere in the APIs where
	a path is needed as argument etc.

	@ivar name: full page name (inherited from L{Path})
	@ivar hascontent: C{True} if the page has content
	@ivar haschildren: C{True} if the page has sub-pages
	@ivar modified: C{True} if the page was modified since the last
	store. Will be reset by L{Notebook.store_page()}
	@ivar readonly: C{True} when the page is read-only or belongs to a readonly notebook

	@signal: C{storage-changed (changed-on-disk)}: signal emitted on page
	change. The argument "changed-on-disk" is C{True} when an external
	edit was detected. For internal edits it is C{False}.
	@signal: C{modified-changed ()}: emitted when the page is edited
	'''

	__signals__ = {
		'storage-changed': (SIGNAL_NORMAL, None, (bool,)),
		'modified-changed': (SIGNAL_NORMAL, None, ()),
	}

	def __init__(self, path, haschildren, file, folder, format):
		assert isinstance(path, Path)
		self.name = path.name
		self.haschildren = haschildren
			# Note: this attribute is updated by the owning notebook
			# when a child page is stored
		self._modified = False
		self._change_counter = 0
		self._parsetree = None
		self._textbuffer = None
		self._meta = None

		self._readonly = None
		self._last_etag = None
		if isinstance(format, str):
			self.format = zim.formats.get_format(format)
		else:
			self.format = format
		self.source_file = file
		self.attachments_folder = folder

	@property
	def readonly(self):
		if self._readonly is None:
			self._readonly = not self.source_file.iswritable()
		return self._readonly

	@property
	def mtime(self):
		return self.source_file.mtime() if self.source_file.exists() else None

	@property
	def ctime(self):
		return self.source_file.ctime() if self.source_file.exists() else None

	@property
	def hascontent(self):
		'''Returns whether this page has content'''
		if self._textbuffer:
			return self._textbuffer.hascontent
		elif self._parsetree:
			return self._parsetree.hascontent
		else:
			return self.source_file.exists()

	@property
	def modified(self):
		return self._modified

	def set_modified(self, modified):
		if modified:
			# HACK: by setting page.modified to a number rather than a
			# bool we can use this number to check against race conditions
			# in notebook.store_page_async post handler
			self._change_counter = max(1, (self._change_counter + 1) % 1000)
			self._modified = self._change_counter
			assert bool(self._modified) is True, 'BUG in counter'
		else:
			self._modified = False
		self.emit('modified-changed')

	def on_buffer_modified_changed(self, buffer):
		# one-way traffic, set page modified after modifying the buffer
		# but do not set page.modified False again when buffer goes
		# back to un-modified. Reason is that we use the buffer modified
		# state to track if we already requested the parse tree (see
		# get_parsetree()) while page modified is used to track need
		# for saving and is reset after save was done
		if buffer.get_modified():
			if self.readonly:
				logger.warning('Buffer edited while page read-only - potential bug')
			self.set_modified(True)

	def _store(self):
		tree = self.get_parsetree()
		self._store_tree(tree)

	def _store_tree(self, tree):
		if tree and tree.hascontent:
			if self._meta is not None:
				tree.meta.update(self._meta) # Preserver headers
			elif self.source_file.exists():
				# Try getting headers from file
				try:
					text = self.source_file.read()
				except zim.newfs.FileNotFoundError:
					return None
				else:
					parser = self.format.Parser()
					tree = parser.parse(text)
					self._meta = tree.meta
					tree.meta.update(self._meta) # Preserver headers
			else: # not self.source_file.exists()
				now = datetime.now()
				tree.meta['Creation-Date'] = now.isoformat()

			lines = self.format.Dumper().dump(tree, file_output=True)
			self._last_etag = self.source_file.writelines_with_etag(lines, self._last_etag)
			self._meta = tree.meta
		else:
			self.source_file.remove()
			self._last_etag = None
			self._meta = None
		self.emit('storage-changed', False)

	def check_source_changed(self):
		'''Checks for changes in the source file and load it if needed

		If the page has a C{textbuffer} and it contains unsaved changes, this
		method will not overwrite them and you'll get an error on next attempt
		to save. To force overwrite see L{reload_textbuffer()}
		'''
		if (
			self._last_etag
			and not (self.source_file.exists() and self.source_file.verify_etag(self._last_etag))
		) or (
			not self._last_etag
			and self.source_file.exists()
		):
			logger.info('Page changed on disk: %s', self.name)
			self._last_etag = None
			self._meta = None
			if self._textbuffer and not self._textbuffer.get_modified():
				self.reload_textbuffer()
			else:
				self._parsetree = None

			self.emit('storage-changed', True)
			return True
		else:
			return False

	def exists(self):
		'''C{True} when the page has either content or children'''
		return self.haschildren or self.hascontent

	def isequal(self, other):
		'''Check equality of pages
		This method is intended to deal with case-insensitive storage
		backends (e.g. case insensitive file system) where the method
		is supposed to check equality of the resource.
		Note that this may be the case even when the page objects differ
		and can have a different name (so L{__cmp__} will not show
		them to be equal). However default falls back to L{__cmp__}.
		@returns: C{True} of both page objects point to the same resource
		@implementation: can be implementated by subclasses
		'''
		if self is other or self == other:
			return True
		elif self.source_file.exists():
			return self.source_file.isequal(other.source_file)
		else:
			return False

	def get_parsetree(self):
		'''Returns the contents of the page

		@returns: a L{zim.formats.ParseTree} object or C{None}
		'''
		if self._textbuffer:
			if self._textbuffer.get_modified() or self._parsetree is None:
				self._parsetree = self._textbuffer.get_parsetree()
				self._textbuffer.set_modified(False)
			#~ print self._parsetree.tostring()
			return self._parsetree
		elif self._parsetree:
			return self._parsetree
		else:
			try:
				text, self._last_etag = self.source_file.read_with_etag()
			except zim.newfs.FileNotFoundError:
				return None
			else:
				parser = self.format.Parser()
				self._parsetree = parser.parse(text, file_input=True)
				self._meta = self._parsetree.meta
				assert self._meta is not None
				return self._parsetree

	def set_parsetree(self, tree):
		'''Set the parsetree with content for this page

		@param tree: a L{zim.formats.ParseTree} object with content
		or C{None} to remove all content from the page

		@note: after setting new content in the Page object it still
		needs to be stored in the notebook to save this content
		permanently. See L{Notebook.store_page()}.
		'''
		if self.readonly:
			raise PageReadOnlyError(self)
		self._set_parsetree(tree)

	def _set_parsetree(self, tree):
		self._parsetree = tree
		if self._textbuffer:
			assert not self._textbuffer.get_modified(), 'BUG: changing parsetree while buffer was changed as well'
			try:
				if tree is None:
					self._textbuffer.clear()
				else:
					self._textbuffer.set_parsetree(tree)
			except:
				# Prevent auto-save to kick in at any cost
				self._textbuffer.set_modified(False)
				raise
			else:
				self._textbuffer.set_modified(False)

		self.set_modified(True)

	def append_parsetree(self, tree):
		'''Append content

		@param tree: a L{zim.formats.ParseTree} object with content
		'''
		if self._textbuffer:
			self._textbuffer.append_parsetree(tree)
		else:
			ourtree = self.get_parsetree()
			if ourtree:
				self.set_parsetree(ourtree + tree)
			else:
				self.set_parsetree(tree)

	def get_textbuffer(self, constructor=None):
		'''Get a C{Gtk.TextBuffer} for the page

		Will either return an existing buffer or construct a new one and return
		it. A C{Gtk.TextBuffer} can be shared between multiple C{Gtk.TextView}s.
		The page object owns the textbuffer to allow multiple views on the same
		page.

		Once a buffer is set, also methods like L{get_parsetree()} and
		L{get_parsetree()} will interact with this buffer.

		@param constructor: if not buffer was set previously, this function
		is called to construct the buffer.

		@returns: a C{TextBuffer} object or C{None} if no buffer is set and
		no constructor is provided.
		'''
		if self._textbuffer is None:
			if constructor is None:
				return None

			tree = self.get_parsetree()
			self._textbuffer = constructor(parsetree=tree)
			self._textbuffer.connect('modified-changed', self.on_buffer_modified_changed)

		return self._textbuffer

	def reload_textbuffer(self):
		'''Reload page content from source file and update the textbuffer if set

			NOTE: this method overwrites any changes in the C{textbuffer} or
			C{parsetree} that have not been saved to file !
		'''
		buffer = self._textbuffer
		self._textbuffer = None
		self._parsetree = None
		if buffer is not None:
			tree = self.get_parsetree()
			self._textbuffer = buffer
			buffer.set_modified(False)
			self._set_parsetree(tree)
				# load new tree in buffer, undo-able in 1 step
				# private method circumvents readonly check !
			self.set_modified(False)
		# else do nothing - source will be read with next call to `get_parsetree()`

	def dump(self, format, linker=None):
		'''Get content in a specific format

		Convenience method that converts the current parse tree to a
		particular format first.

		@param format: either a format module or a string
		that is understood by L{zim.formats.get_format()}.

		@param linker: a linker object (see e.g. L{BaseLinker})

		@returns: text as a list of lines or an empty list
		'''
		if isinstance(format, str):
			format = zim.formats.get_format(format)

		if not linker is None:
			linker.set_path(self)

		tree = self.get_parsetree()
		if tree:
			return format.Dumper(linker=linker).dump(tree)
		else:
			return []

	def parse(self, format, text, append=False):
		'''Store formatted text in the page

		Convenience method that parses text and sets the parse tree
		accordingly.

		@param format: either a format module or a string
		that is understood by L{zim.formats.get_format()}.
		@param text: text as a string or as a list of lines
		@param append: if C{True} the text is appended instead of
		replacing current content.
		'''
		if isinstance(format, str):
			format = zim.formats.get_format(format)

		if append:
			self.append_parsetree(format.Parser().parse(text))
		else:
			self.set_parsetree(format.Parser().parse(text))

	def get_title(self):
		tree = self.get_parsetree()
		if tree:
			return tree.get_heading_text() or self.basename
		else:
			return self.basename

	def heading_matches_pagename(self):
		'''Returns whether the heading matches the page name.
		Used to determine whether the page should have its heading
		auto-changed on rename/move.
		@returns: C{True} when the heading can be auto-changed.
		'''
		tree = self.get_parsetree()
		if tree:
			return tree.get_heading_text() == self.basename
		else:
			return False
